/*
 * VXPLib COMPONENT
 *
 * Module: GlobalObject.cpp - Implementation of some global classes of
 * the component
 *
 * Synopsis: Find here implementations of a few classes whose secondary role
 * is in providing extra functionality and being information holders
 * for other clases and logical objects of the component.
 *
 * Copyright (c) 2002, Vitaly Tomilov
 * Copyright (c) 2003-2005, Tooltips.Net L.L.C.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * Library support web site - http://www.tooltips.net
 */

#include "stdafx.h"
#include "GlobalObject.h"
#include "VXPLib.h"

// *************************************************
// CGradientInfo implementation ********************
// *************************************************

CGradientInfo::CGradientInfo()
{
	Reset();
}

void CGradientInfo::Reset()
{
	m_bGFCHorizontal = true;
	m_bGFCValid = false;
}

void CGradientInfo::Init(CTag * pTag)
{
	CString grad = pTag->GetAttrTextValue(_T("GRAD"));
	if(!grad.IsEmpty())
	{
		grad.Remove(' ');
		grad.MakeUpper();
		if(!grad.IsEmpty() && grad == _T("VERT"))
			m_bGFCHorizontal = false;
	}
	m_GFC = pTag->GetAttrColorValue(_T("GFC"), m_bGFCValid);
}

// *************************************************
// CFontObject implementation **********************
// *************************************************

CFontObject::CFontObject(LOGFONTEX * lf)
{
	m_Font = NULL;
	m_LogFont = *lf;
}

CFontObject::~CFontObject()
{
	if(m_Font)
		DeleteObject(m_Font);
}

HGDIOBJ CFontObject::SelectFont(HDC dc, bool IsBold, bool IsItalic, bool IsUnderlined, bool IsStriked)
{
	if(m_Font)
		DeleteObject(m_Font);
	LOGFONTEX lf = m_LogFont;
	if(IsBold && lf.lfWeight < FW_BOLD)
		lf.lfWeight = FW_BOLD;
	if(IsItalic)
		lf.lfItalic = TRUE;
	if(IsUnderlined)
		lf.lfUnderline = TRUE;
	if(IsStriked)
		lf.lfStrikeOut = TRUE;
	m_Font = CreateFontIndirect(&lf);
	SetTextColor(dc, lf.m_TextColor);
	
	HGDIOBJ selected = SelectObject(dc, m_Font);

	if(m_LogFont.m_bBkColorValid)
	{
		SetBkMode(dc, OPAQUE);
		SetBkColor(dc, m_LogFont.m_TextBkColor);
	}
	else
		SetBkMode(dc,TRANSPARENT);
	return selected;
}

void CFontObject::GetFont(LOGFONTEX *lf)
{
	*lf = m_LogFont;
}

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

CGlobalObject::CGlobalObject():m_pFont(NULL)
{
	m_pLastRequestedControl = NULL;
	m_LastRequestedIndex = 0;

	m_bNewLine = true;

	m_BodyCodePage = CP_ACP;	// ANSI Code Page
	m_TxtCodePage = CP_ACP;		// ANSI Code Page

	m_pFontList = NULL;
	m_DefaultFont = NULL;
	m_pPathList = NULL;
	m_hDefaultRes = NULL;

	ResetState();

// *** Setting up the default resource file ***
	CString path;
	GetModuleFileName(NULL, path.PreAllocate(MAX_PATH_LENGTH), MAX_PATH_LENGTH);
	path.PostAllocate();

#ifdef _MSLU
	if(g_bNTCompitable)
		m_hDefaultRes = ::LoadLibraryEx(path, NULL, LOAD_LIBRARY_AS_DATAFILE);
	else
		m_hDefaultRes = ::LoadLibrary(path);
#else
	m_hDefaultRes = ::LoadLibraryEx(path, NULL, LOAD_LIBRARY_AS_DATAFILE);
#endif
	

// *** Adding the default directory ***********
	int index = path.ReverseFind('\\');
	if(index != -1)
		ADD_COLLECTION(m_pPathList, new CPathList(path.Left(index)))

// *** Adding path to the component as another resource folder

	GetModuleFileName(GetModuleHandle(_T("VXPLib.DLL")), path.PreAllocate(MAX_PATH_LENGTH), MAX_PATH_LENGTH);
	path.PostAllocate();
	index = path.ReverseFind('\\');
	if(index != -1)
		ADD_COLLECTION(m_pPathList, new CPathList(path.Left(index)))
}

CGlobalObject::~CGlobalObject()
{
	if(m_hDefaultRes)
		FreeLibrary(m_hDefaultRes);
	if(m_DefaultFont)
		DeleteObject(m_DefaultFont);

	DELETE_COLLECTION(m_pFontList)
	DELETE_COLLECTION(m_pPathList)
}

HGDIOBJ CGlobalObject::SelectCurrentFont(HDC dc)
{
	if(m_pFontList)
		return m_pFontList->SelectFont(dc, m_uBoldCounter > 0, m_uItalicCounter > 0, m_uUnderlineCounter > 0, m_uStrikeCounter > 0);

	if(m_DefaultFont)
		DeleteObject(m_DefaultFont);
	LOGFONTEX lf = m_DefaultLogFont;
	if(m_uBoldCounter > 0 && lf.lfWeight < FW_BOLD)
		lf.lfWeight = FW_BOLD;
	if(m_uItalicCounter > 0)
		lf.lfItalic = TRUE;
	if(m_uUnderlineCounter > 0)
		lf.lfUnderline = TRUE;
	if(m_uStrikeCounter > 0)
		lf.lfStrikeOut = TRUE;
	m_DefaultFont = CreateFontIndirect(&lf);
	SetTextColor(dc, m_DefaultLogFont.m_TextColor);
	if(m_DefaultLogFont.m_bBkColorValid)
	{
		SetBkMode(dc, OPAQUE);
		SetBkColor(dc, m_DefaultLogFont.m_TextBkColor);
	}
	else
		SetBkMode(dc,TRANSPARENT);
	return SelectObject(dc, m_DefaultFont);
}

void CGlobalObject::GetCurrentFont(LOGFONTEX *lf)
{
	if(m_pFontList)
		m_pFontList->GetFont(lf);
	else
		*lf = m_DefaultLogFont;
}

void CGlobalObject::AddFont(LOGFONTEX *lf)
{
	CFontObject * font = new CFontObject(lf);
	if(m_pFontList)
		m_pFontList->Add(font);
	m_pFontList = font;
}

void CGlobalObject::SetPreviousFont()
{
	if(!m_pFontList)
		return;
	if(m_pFontList->Prev)
	{
		CFontObject * prev = (CFontObject*)m_pFontList->Prev;
		m_pFontList->Remove();
		m_pFontList = prev;
	}
	else
	{
		delete m_pFontList;
		m_pFontList = NULL;
	}
}

void CGlobalObject::ResetState()
{
	DELETE_COLLECTION(m_pFontList)
	
	m_bNewLine = true;

	m_pLastRequestedControl = NULL;
	m_LastRequestedIndex = 0;

	m_grad.Reset();

	// Reset all counters
	m_uBoldCounter = 0;
	m_uItalicCounter = 0;
	m_uUnderlineCounter = 0;
	m_uStrikeCounter = 0;

	if(!m_pFont)
		return;

	if(m_DefaultFont)
		DeleteObject(m_DefaultFont);

	long font_handle;
	m_pFont->get_hFont(&font_handle);
	GetObject((HGDIOBJ)font_handle, sizeof(LOGFONT), &m_DefaultLogFont);
	short transparent;
	m_pFont->get_Transparent(&transparent);
	m_DefaultLogFont.m_bBkColorValid = (transparent == 0);
	m_pFont->get_Color(&m_DefaultLogFont.m_TextColor);
	m_pFont->get_BackColor(&m_DefaultLogFont.m_TextBkColor);
	m_DefaultFont = CreateFontIndirect(&m_DefaultLogFont);
}

LOGFONTEX CGlobalObject::GetDefaultLogFont()
{
	return m_DefaultLogFont;
}

void CGlobalObject::IncrementBold()
{
	m_uBoldCounter ++;
}

void CGlobalObject::DecrementBold()
{
	m_uBoldCounter --;
	if(m_uBoldCounter < 0)
		m_uBoldCounter = 0;
}

void CGlobalObject::IncrementItalic()
{
	m_uItalicCounter ++;
}

void CGlobalObject::DecrementItalic()
{
	m_uItalicCounter --;
	if(m_uItalicCounter < 0)
		m_uItalicCounter = 0;
}

void CGlobalObject::IncrementUnderline()
{
	m_uUnderlineCounter ++;
}

void CGlobalObject::DecrementUnderline()
{
	m_uUnderlineCounter --;
	if(m_uUnderlineCounter < 0)
		m_uUnderlineCounter = 0;
}

void CGlobalObject::IncrementStrike()
{
	m_uStrikeCounter ++;
}

void CGlobalObject::DecrementStrike()
{
	m_uStrikeCounter --;
	if(m_uStrikeCounter < 0)
		m_uStrikeCounter = 0;
}

void CGlobalObject::TextOut(HDC dc, int x, int y, CString & txt)
{
	TabbedTextOut(dc, x, y, txt, txt.GetLength(), 0, NULL, x);
}

CSize CGlobalObject::GetTextSize(HDC dc, CString &txt)
{
	DWORD size = GetTabbedTextExtent(dc, txt, txt.GetLength(), 0, NULL);
	return CSize(LOWORD(size), HIWORD(size));
}

// Adds a new file path into the list and returns
// the length of the list.
short CGlobalObject::AddFilePath(BSTR FilePath)
{
	ADD_COLLECTION(m_pPathList, new CPathList(FilePath))
	return m_pPathList->CountAll();
}

// This method must make sure that the returned file path
// is correct. Otherwise it must return an empty string.
CString CGlobalObject::GetValidFilePath(LPCTSTR FileName, HANDLE * hFile)
{
	CString result;

	if(!FileName)
		return result;

	if(FileExists(FileName, hFile))
		return FileName;
	CPathList * plist = m_pPathList;
	CString tmp;
	while(plist)
	{
		tmp = plist->m_FilePath;
		tmp += '\\';
		tmp += FileName;
		if(FileExists(tmp, hFile))
			return tmp;
		plist = (CPathList*)plist->Next;
	}
	return result;
}

HINSTANCE CGlobalObject::GetDefaultInstance()
{
	return m_hDefaultRes;
}

CSize CGlobalObject::GetDefIconSize()
{
	return CSize(0, 0);
}

void CGlobalObject::SetDefaultRes(BSTR FilePath, VARIANT_BOOL * succeeded)
{
	CString path(FilePath);
	// Try on all paths:
	path = GetValidFilePath(path);
	// Assume failure
	*succeeded = FALSE;
	if(path.IsEmpty())
		return;

#ifdef _MSLU
	HINSTANCE hDefaultRes;
	if(g_bNTCompitable)
		hDefaultRes = ::LoadLibraryEx(path, NULL, LOAD_LIBRARY_AS_DATAFILE);
	else
		hDefaultRes = ::LoadLibrary(path);
#else
	HINSTANCE hDefaultRes = ::LoadLibraryEx(path, NULL, LOAD_LIBRARY_AS_DATAFILE);
#endif

	if(hDefaultRes)
	{
		*succeeded = TRUE;
		if(m_hDefaultRes)
			FreeLibrary(m_hDefaultRes);
		m_hDefaultRes = hDefaultRes;
	}
}

bool CGlobalObject::IsItalic()
{
	return (m_uItalicCounter || m_DefaultLogFont.lfItalic);
}

long CGlobalObject::GetTxtCodePage()
{
	return m_TxtCodePage;
}
