/*
 * VXPLib COMPONENT
 *
 * Module: HCollection.h - Declaration of class HCollection
 *
 * Synopsis: Class HCollection implements a colection of abstract
 * elements. Any type derived from this class will gain properties of
 * a collection element.
 *
 * Copyright (c) 2002, Vitaly Tomilov
 * Copyright (c) 2003-2005, Tooltips.Net L.L.C.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * Library support web site - http://www.tooltips.net
 */

#ifndef HLib_Collection
#define HLib_Collection

// This macro simplifies adding elements in collection
#define ADD_COLLECTION(A,B)\
{\
	if (A)\
		A->Add(B);\
	else\
		A = B;\
}

// This macro simplifies including elements in collection
#define INCLUDE_COLLECTION(A,B)\
{\
	if (A)\
		A->Include(B);\
	else\
		A = B;\
}

// This macro simplifies deleting alements from collection
#define DELETE_COLLECTION(A)\
{\
	if (A)\
	{\
		delete A;\
		A = 0;\
	};\
}

// Function type to be used when calling ForEach method
// Parameters: caller, receiver, parameter
typedef void (*CallFunction)(void*,void*,void*);

//*** Declaration of the class itself **************************

class HCollection
{
public:
// This connection object is used when you need to single out an object from
// the collection by calling Cut(), but eventually, you do not want to destroy
// the connection, then you just memorize the connection by calling
// HCollection::GetConnection(); When you are done with this single object,
// in order to put it back in the collection simply call HCollection::SetConnection()
// with the parameter, returned to you by HCollection::GetConnection().
struct connection
{
	connection(HCollection * prev, HCollection * next);
	HCollection * Prev;
	HCollection * Next;
};

	HCollection *Prev; // Pointer to the previous element in collection
	HCollection *Next; // Pointer to the next element in collection

// Beware of modifying the collection elements between calling
// GetConnection() and SetConnection, as it may result in wrong
// reconnection of the object, and memory corrupt
	connection GetConnection(); // Save the connection (Normally, before calling Cut())
	bool SetConnection(connection); // Restore the connection

	HCollection(); // Default constructor
	HCollection(HCollection*);	// Constructor which attaches the collection-
								// -parameter to its own collection

	virtual ~HCollection();	// Virtual destructor

// Get the very first object in this collection
	HCollection * GetFirst();

// Get the very last object in this collection
	HCollection * GetLast();

// Append a collection-parameter to this collection
// Return value - the parameter passed
	HCollection * Add(HCollection*);

// Insert a collection of objects right after
// this object
// Return value - the parameter passed
	HCollection * Insert(HCollection*);

// Include a collection of objects in this one and sort it,
// assuming this collection has been sorted before
// Return value - the parameter passed
	HCollection * Include(HCollection*);

// Cut this object from the collection chain
// and return it
	HCollection * Cut();

// Split all objects in the collection from each other
// (break all connections between them)
	void CutAll();

// Remove this object from the collection and destroy it
// Analogue to: delete Cut();
	void Remove();

// Make the first item of collection the last one
	void FirstLast();

// Make the last item of collection the first one
	void LastFirst();

// Make this object on the first place in collection
	void MoveFirst();

// Make this object on the last place in collection
	void MoveLast();

// Sort the entire collection depending on how the
// method Compare(...) overloaded.
	void Sort();

// Reverse the entire collection
	void Reverse();

// Get index of this object in the collection
	int GetIndex();

// Get the number of all elements in the collection
	int CountAll();

// Get the number of all items in the collection
// before this object
	int CountBefore();

// Get the number of all items in the collection
// after this object
	int CountAfter();

// Get an object from the collection by its index,
// starting with 0.
// If index is out of range it returns 0.
	HCollection * GetObject(int Index);

// Call a template method for every item in the collection
// with parameters: caller, receiver, parameter
	void ForEach(CallFunction func, void * parameter = 0);

// Find an object in the collection connected with this object
// If no such object found, this object is returned
// It checks: Prev, Next, then This.
	HCollection * GetChain();

// Make sure this object was sorted in the collection
// If not, then sort it
	void CheckSorted();

// This operator appends the collection-parameter
// and returns its last element
	HCollection * operator +  (HCollection*);

// This operator appends the collection-parameter
// and returns the parameter itself
	HCollection * operator += (HCollection*);

protected:

// Compare this object with the object-parameter
	virtual bool Compare(HCollection*);

private:

// Get the very extremal object (Min or Max) using a comparison template
	HCollection * GetLim();
};

template<class T>
class TCollection:public HCollection, public T{};

#endif
