/*
 * VXPLib COMPONENT
 *
 * Module: Picture.cpp - Image library
 *
 * Synopsis: This file contains implementations of image classes
 * of some of the most known graphical formats: ICO, CUR, ANI, BMP, GIF,
 * JPG, EMF and WMF.
 *
 * Copyright (c) 2002, Vitaly Tomilov
 * Copyright (c) 2003-2005, Tooltips.Net L.L.C.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * Library support web site - http://www.tooltips.net
 */

#include "stdafx.h"
#include "Picture.h"

// This list of all supported graphic formats:
#define RT_NONE		0	// Wrong format
#define RT_CUSTOM	1	// Custom type name
#define RT_BMP		2	// BMP type name...
#define RT_ICO		3
#define RT_CUR		4
#define RT_ANI		5
#define RT_GIF		6
#define RT_JPG		7
#define RT_EMF		8
#define RT_WMF		9

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

CSize CPicture::GetSize()
{
	return m_size;
}

void CPicture::SetSize(int cx, int cy)
{
	m_size.SetSize(cx, cy);
}

void CPicture::PopulateWithBitmap(HDC dc, HBITMAP bmp, int x, int y, int cx, int cy)
{
	if(m_size.IsEmpty())
		return;

	HDC memDC = CreateCompatibleDC(dc);
	HGDIOBJ old = SelectObject(memDC, bmp);

	int i;

	// Combinations of the last parameter SRCCOPY can produce
	// many interesting effects
		
	// Step1: Filling most of the area
	for(i = 0;i < cy / m_size.cy;i ++)
		for(int k = 0;k < cx / m_size.cx;k ++)
			BitBlt(dc, x + k * m_size.cx, y + i * m_size.cy, m_size.cx, m_size.cy, memDC, 0, 0, SRCCOPY);
		
	// Step2: Filling the remainder on the right:
	int Shift_X = ((int)cx / m_size.cx) * m_size.cx;
	for(i = 0;i < cy / m_size.cy;i ++)
		BitBlt(dc, x + Shift_X, y + i * m_size.cy, cx - Shift_X, m_size.cy, memDC, 0, 0, SRCCOPY);

	// Step3: Filling the remainder on the bottom:
	int Shift_Y = ((int)cy / m_size.cy) * m_size.cy;
	for(i = 0;i < cx / m_size.cx;i ++)
		BitBlt(dc, x + i * m_size.cx, y + Shift_Y, m_size.cx, cy - Shift_Y, memDC, 0, 0, SRCCOPY);

	// Step4: Filling the right bottom corner
	BitBlt(dc, x + Shift_X, y + Shift_Y, cx - Shift_X, cy - Shift_Y, memDC, 0, 0, SRCCOPY);

	SelectObject(memDC, old);
	DeleteDC(memDC);
}

void CPicture::SetBitmapDimensions(HBITMAP bmp, int cxDesired, int cyDesired)
{
	m_size.cx = cxDesired;
	m_size.cy = cyDesired;

	if(!cxDesired || !cyDesired)
	{
		BITMAPINFO bmpInfo;
		bmpInfo.bmiHeader.biSize = sizeof(BITMAPINFOHEADER);
		bmpInfo.bmiHeader.biBitCount = 0;
		HDC dc = GetDC(NULL);
		if(GetDIBits(dc, bmp, 0, 0, NULL, &bmpInfo, DIB_RGB_COLORS))
		{
			if(!cxDesired)
				m_size.cx = bmpInfo.bmiHeader.biWidth;
			if(!cyDesired)
				m_size.cy = bmpInfo.bmiHeader.biHeight;
		}
		ReleaseDC(NULL, dc);
	}
}

// When no known file extension is found it tries the IPicture
// interface to recognize the file format and display it.
CPicture * CPicture::GetObjectFromFile(LPCTSTR file, CGlobalObject * globalObject, int cxDesired, int cyDesired)
{
	CString txt = globalObject->GetValidFilePath(file);
	if(txt.IsEmpty())
		return NULL;
	txt.TrimRight();
	CString FullPath = txt;

	int index = txt.ReverseFind('.');
	int L = txt.GetLength();
	
	// For now we do not support file extensions of the
	// length more than 4 symbols, i.e. an extension like JPEG
	// is supported, but not longer than that.
	if(index == -1 || index < (L - 4) || (index + 1) == L)
		return NULL;
	txt = txt.Right(L - index - 1);
	txt.MakeUpper();
	if(txt == _T("ICO"))
	{
		CIcon * icon = new CIcon;
		if(!cxDesired || !cyDesired)
		{
			CSize s = globalObject->GetDefIconSize();
			if(!cxDesired)
				cxDesired = s.cx;
			if(!cyDesired)
				cyDesired = s.cy;
		}
		if(icon->LoadFromFile(FullPath, cxDesired, cyDesired))
			return icon;
		delete icon;
		return NULL;
	}
	if(txt == _T("CUR") || txt == _T("ANI"))
	{
		CIcon * icon = new CIcon(true);
		if(icon->LoadFromFile(FullPath, cxDesired, cyDesired))
			return icon;
		delete icon;
		return NULL;
	}
	if(txt == _T("BMP"))
	{
		CBmp * bmp = new CBmp;
		if(bmp->LoadFromFile(FullPath, cxDesired, cyDesired))
			return bmp;
		delete bmp;
		return NULL;
	}

	// If none found so far, then we forward all following
	// attempts to the IPicture interface
	if(txt == _T("JPG") || txt == _T("JPEG") || txt == _T("GIF") || txt == _T("EMF") || txt == _T("WMF"))
	{
		CComPicture * pic = new CComPicture;
		if(pic->LoadFromFile(FullPath, cxDesired, cyDesired))
			return pic;
		delete pic;
	}
	return NULL;
}

CPicture * CPicture::GetObjectFromTag(CTag *pTag, LPCTSTR SourceName, CGlobalObject *globalObject, int cxDesired, int cyDesired)
{
	bool bSourceSpecified, bResSpecified;

	CString source, FileSource;
	if(SourceName)
		FileSource = pTag->GetAttrTextValue(SourceName, &bSourceSpecified);
	else
	{
		FileSource = pTag->GetAttrTextValue(attrBG, &bSourceSpecified);
		if(!bSourceSpecified)
			FileSource = pTag->GetAttrTextValue(attrBACKGROUND, &bSourceSpecified);
	}
	
	source = FileSource;

	CString Res = pTag->GetAttrTextValue(attrRES, &bResSpecified);

	if(bResSpecified)
	{
		if(bSourceSpecified)
		{
			if(source.IsEmpty())
				return NULL;
			source = globalObject->GetValidFilePath(source);
			if(source.IsEmpty())
				return NULL;


		#ifdef _MSLU
			HINSTANCE hInst;
			if(g_bNTCompitable)
				hInst = ::LoadLibraryEx(source, NULL, LOAD_LIBRARY_AS_DATAFILE);
			else
				hInst = ::LoadLibrary(source);
		#else
			HINSTANCE hInst = ::LoadLibraryEx(source, NULL, LOAD_LIBRARY_AS_DATAFILE);
		#endif
			
			if(!hInst)
				return NULL;
			CPicture * result = GetObjectFromResource(globalObject, hInst, Res, cxDesired, cyDesired);
			FreeLibrary(hInst);
			return result;
		}
		HINSTANCE hDefInst = globalObject->GetDefaultInstance();
		if(!hDefInst)
			return NULL;
		return GetObjectFromResource(globalObject, hDefInst, Res, cxDesired, cyDesired);
	}
	if(bSourceSpecified)
	{
		if(FileSource.IsEmpty())
			return NULL;
		return GetObjectFromFile(FileSource, globalObject, cxDesired, cyDesired);
	}

	return NULL;
}

CPicture * CPicture::GetObjectFromResource(CGlobalObject * globalObject, HINSTANCE hInst, CString Res, int cxDesired, int cyDesired)
{
	CString TypeName, ResName;
	int rt = GetResInfo(Res, TypeName, ResName);
	if(rt == RT_NONE)
		return NULL;

// Step1: If it is one of the standard resources (BMP, ICO, CUR, ANI),
// then we'll first try to load them the standard way.

	if(rt == RT_BMP)
	{
		CBmp * bmp = new CBmp;
		if(bmp->LoadFromRes(hInst, ResName, cxDesired, cyDesired))
			return bmp;
		delete bmp;
	}
	if(rt == RT_ICO)
	{
		CIcon * icon = new CIcon;
		if(!cxDesired || !cyDesired)
		{
			CSize s = globalObject->GetDefIconSize();
			if(!cxDesired)
				cxDesired = s.cx;
			if(!cyDesired)
				cyDesired = s.cy;
		}
		if(icon->LoadFromRes(hInst, ResName, cxDesired, cyDesired))
			return icon;
		delete icon;
	}
	if(rt == RT_CUR || rt == RT_ANI)
	{
		CIcon * icon = new CIcon(true);
		if(icon->LoadFromRes(hInst, ResName, cxDesired, cyDesired))
			return icon;
		delete icon;
	}

// Step2: Since we failed to load the resource the standard way,
// we load it as a custom resource using the CComPicture class only.

	CComPicture * pic = new CComPicture;
	if(pic->LoadFromRes(hInst, TypeName, ResName, cxDesired, cyDesired))
		return pic;
	delete pic;
	return NULL;
}

void CPicture::ResetSize()
{
	m_size.Zero();
}

// Formats:
// 1: "BMP=#142"
// 2: "BMP=IDR_MYBMP"
// 3: "MYRESTYPE=IDR_MYPIC"
// Returns the resource type, and sets
// the resource type name (if RT_) and the resource name:
int CPicture::GetResInfo(CString res, CString & ResType, CString & ResName)
{
	res.Remove(' '); // Cleaning from white spaces
	int L = res.GetLength();
	int index = res.Find('=');
	if(index == -1)
		index = res.Find(':');
	if(index < 1 || L - index < 2)
		return RT_NONE;
	ResType = res.Left(index);
	ResType.MakeUpper();
	ResName = res.Right(L - index - 1);
	ResName.MakeUpper();
	if(ResType.GetLength() != 3)
		return RT_CUSTOM;
	if(ResType == _T("BMP"))
		return RT_BMP;
	if(ResType == _T("ICO"))
		return RT_ICO;
	if(ResType == _T("CUR"))
		return RT_CUR;
	if(ResType == _T("ANI"))
		return RT_ANI;
	if(ResType == _T("GIF"))
		return RT_GIF;
	if(ResType == _T("EMF"))
		return RT_EMF;
	if(ResType == _T("WMF"))
		return RT_WMF;
	return RT_CUSTOM;
}

/////////////////////////////////////////////
// CIcon class
/////////////////////////////////////////////

CIcon::CIcon(bool IsCursor)
{
	m_bCursor = IsCursor;
	m_hIcon = NULL;
	m_IconInfo.hbmColor = NULL;
	m_IconInfo.hbmMask = NULL;
}

CIcon::~CIcon()
{
	UnloadIcon();
}

bool CIcon::LoadFromRes(HINSTANCE hInst, CString ResID, int cxDesired, int cyDesired)
{
	UnloadIcon();
	m_hIcon = (HICON)LoadImage(hInst, ResID, m_bCursor?IMAGE_CURSOR:IMAGE_ICON, cxDesired, cyDesired, LR_DEFAULTSIZE | LR_DEFAULTCOLOR);
	if(!m_hIcon)
		return false;
	GetIconInfo(m_hIcon, &m_IconInfo);
	SetBitmapDimensions(m_IconInfo.hbmColor, cxDesired, cyDesired);
	return true;
}

bool CIcon::LoadFromFile(LPCTSTR FileName, int cxDesired, int cyDesired)
{
	UnloadIcon();
	m_hIcon = (HICON)LoadImage(NULL, FileName, m_bCursor?IMAGE_CURSOR:IMAGE_ICON, cxDesired, cyDesired, LR_DEFAULTSIZE | LR_LOADFROMFILE);
	if(!m_hIcon)
		return false;
	GetIconInfo(m_hIcon, &m_IconInfo);
	SetBitmapDimensions(m_IconInfo.hbmColor, cxDesired, cyDesired);
	return true;
}

void CIcon::UnloadIcon()
{
	ResetSize();
	if(m_hIcon)
	{
		DestroyIcon(m_hIcon);
		m_hIcon = NULL;
		if(m_IconInfo.hbmColor)
		{
			DeleteObject(m_IconInfo.hbmColor);
			m_IconInfo.hbmColor = NULL;
		}
		if(m_IconInfo.hbmMask)
		{
			DeleteObject(m_IconInfo.hbmMask);
			m_IconInfo.hbmMask = NULL;
		}
	}
}

// Draw single icon
void CIcon::Draw(HDC dc, int x, int y, int cx, int cy)
{
	if(!m_hIcon)
		return;

	if(cx < 1 || cx > m_size.cx)
		cx = m_size.cx;
	if(cy < 1 || cy > m_size.cy)
		cy = m_size.cy;

	::DrawIconEx(dc, x, y, m_hIcon, cx, cy, 0, NULL, DI_NORMAL);
}

void CIcon::Populate(HDC dc, int x, int y, int cx, int cy)
{
	if(!m_hIcon)
		return;

	HDC MemDC = CreateCompatibleDC(dc);
	int SavedMemDC = SaveDC(MemDC);
	HBITMAP MemBmp = CreateCompatibleBitmap(dc, cx, cy);
	HGDIOBJ OldObject = SelectObject(MemDC, MemBmp);
	BitBlt(MemDC, 0, 0, cx, cy, dc, x, y, SRCCOPY);

	for(int i = 0;i <= cy / m_size.cy;i ++)
		for(int k = 0;k <= cx / m_size.cx;k ++)
			::DrawIconEx(MemDC, k * m_size.cx, i * m_size.cy, m_hIcon, m_size.cx, m_size.cy, 0, NULL, DI_NORMAL);

	BitBlt(dc, x, y, cx, cy, MemDC, 0, 0, SRCCOPY);
	SelectObject(MemDC, OldObject);
	RestoreDC(MemDC, SavedMemDC);
	DeleteObject(MemBmp);
	DeleteDC(MemDC);
}

/////////////////////////////////////////////
// CBmp class
/////////////////////////////////////////////

CBmp::CBmp()
{
	m_hBitmap = NULL;
}

CBmp::~CBmp()
{
	UnloadBitmap();
}

bool CBmp::LoadFromRes(HINSTANCE hInst, CString ResID, int cxDesired, int cyDesired)
{
	UnloadBitmap();
	m_hBitmap = (HBITMAP)LoadImage(hInst, ResID, IMAGE_BITMAP, cxDesired, cyDesired, LR_DEFAULTSIZE | LR_DEFAULTCOLOR);
	if(m_hBitmap)
	{
		SetBitmapDimensions(m_hBitmap, cxDesired, cyDesired);
		return true;
	}
	return false;
}

bool CBmp::LoadFromFile(LPCTSTR FileName, int cxDesired, int cyDesired)
{
	UnloadBitmap();
	m_hBitmap = (HBITMAP)LoadImage(NULL, FileName, IMAGE_BITMAP, cxDesired, cyDesired, LR_DEFAULTSIZE | LR_LOADFROMFILE);
	if(m_hBitmap)
	{
		SetBitmapDimensions(m_hBitmap, cxDesired, cyDesired);
		return true;
	}
	return false;
}

void CBmp::UnloadBitmap()
{
	ResetSize();
	if(m_hBitmap)
	{
		DeleteObject(m_hBitmap);
		m_hBitmap = NULL;
	}
}

// Draw single icon
void CBmp::Draw(HDC dc, int x, int y, int cx, int cy)
{
	if(m_hBitmap)
	{
		if(cx < 1 || cx > m_size.cx)
			cx = m_size.cx;
		if(cy < 1 || cy > m_size.cy)
			cy = m_size.cy;
		HDC memDC = CreateCompatibleDC(dc);
		HGDIOBJ old = SelectObject(memDC, m_hBitmap);
		BitBlt(dc, x, y, cx, cy, memDC, 0, 0, SRCCOPY);
		SelectObject(memDC, old);
		DeleteDC(memDC);
	}
}

void CBmp::Populate(HDC dc, int x, int y, int cx, int cy)
{
	if(m_hBitmap)
		PopulateWithBitmap(dc, m_hBitmap, x, y, cx, cy);
}

/////////////////////////////////////////////
// CComPicture class
/////////////////////////////////////////////

CComPicture::CComPicture()
{
	m_pPicture = NULL;
}

CComPicture::~CComPicture()
{
	UnloadPicture();
}

bool CComPicture::LoadFromRes(HINSTANCE hInst, CString TypeName, CString ResName, int cxDesired, int cyDesired)
{
	UnloadPicture();
	bool bResult = false;
	BYTE * pBuff = NULL;
	int nSize = 0;
	if(GetResource(hInst, ResName, TypeName, pBuff, nSize))
	{
		if(nSize < 1)
			return false;
		pBuff = new BYTE[nSize];
		if(GetResource(hInst, ResName, TypeName, pBuff, nSize))
		{
			if(LoadFromBuffer(pBuff, nSize, cxDesired, cyDesired))
				bResult = true;
		}
		delete [] pBuff;
	}
	return bResult;
}

bool CComPicture::LoadFromFile(LPCTSTR FileName, int cxDesired, int cyDesired)
{
	UnloadPicture();
	int size;
	BYTE * buffer = ReadFromFile(FileName, size);
	if(!buffer)
		return false;
	bool result = LoadFromBuffer(buffer, size, cxDesired, cyDesired);
	delete []buffer;
	return result;
}

#define HIMETRIC_INCH 2540

bool CComPicture::LoadFromBuffer(BYTE * buffer, DWORD nSize, int cxDesired, int cyDesired)
{
	bool bResult = false;
	HGLOBAL hGlobal = GlobalAlloc(GMEM_MOVEABLE, nSize);
	void * pData = GlobalLock(hGlobal);
	memcpy(pData, buffer, nSize);
	GlobalUnlock(hGlobal);
	IStream * pStream = NULL;
	if(CreateStreamOnHGlobal(hGlobal, TRUE, &pStream) == S_OK)
	{
		HRESULT hr;
		if((hr = OleLoadPicture(pStream, nSize, FALSE, IID_IPicture, (LPVOID *)&m_pPicture)) == S_OK)
		{
			bResult = true;
			HDC dc = NULL;
			if(cxDesired > 0)
				m_size.cx = cxDesired;
			else
			{
				dc = GetDC(NULL);
				long width;
				m_pPicture->get_Width(&width);
				width = MulDiv(width, GetDeviceCaps(dc, LOGPIXELSX), HIMETRIC_INCH);
				m_size.cx = width;
			}
			if(cyDesired > 0)
				m_size.cy = cyDesired;
			else
			{
				if(!dc)
					dc = GetDC(NULL);
				long height;
				m_pPicture->get_Height(&height);
				height = MulDiv(height, GetDeviceCaps(dc, LOGPIXELSY), HIMETRIC_INCH);
				m_size.cy = height;
			}
			if(dc)
				ReleaseDC(NULL, dc);
		}
		pStream->Release();
	}
	return bResult;
}

void CComPicture::UnloadPicture()
{
	ResetSize();
	if (m_pPicture != NULL)
	{
		m_pPicture->Release();
		m_pPicture = NULL;
	}
}

void CComPicture::Draw(HDC dc, int x, int y, int cx, int cy)
{
	if(m_pPicture)
	{
		if(cx < 1 || cx > m_size.cx)
			cx = m_size.cx;
		if(cy < 1 || cy > m_size.cy)
			cy = m_size.cy;
		long hmWidth;
		long hmHeight;
		m_pPicture->get_Width(&hmWidth);
		m_pPicture->get_Height(&hmHeight);
		hmWidth = hmWidth * cx/m_size.cx;
		int nCopy = hmHeight * cy/m_size.cy;
		m_pPicture->Render(dc, x, y, cx, cy, 0, hmHeight, hmWidth, -nCopy, NULL);
	}
}

void CComPicture::Populate(HDC dc, int x, int y, int cx, int cy)
{
	if(!m_pPicture)
		return;

	long hmWidth;
	long hmHeight;
	m_pPicture->get_Width(&hmWidth);
	m_pPicture->get_Height(&hmHeight);

	int i, k;

	// Step1: Filling most of the area
	for(i = 0;i < cy / m_size.cy;i ++)
		for(k = 0;k < cx / m_size.cx;k ++)
			m_pPicture->Render(dc, x + k * m_size.cx, y + i * m_size.cy, m_size.cx, m_size.cy, 0, hmHeight, hmWidth, -hmHeight, NULL);
	
	// Step2: Filling the remainder on the right:
	int Shift_X = ((int)cx / m_size.cx) * m_size.cx;
	for(i = 0;i < cy / m_size.cy;i ++)
		m_pPicture->Render(dc, x + Shift_X, y + i * m_size.cy, cx - Shift_X, m_size.cy, 0, hmHeight, hmWidth * (cx - Shift_X)/m_size.cx, -hmHeight, NULL);
	
	// Step3: Filling the remainder on the bottom:
	int Shift_Y = ((int)cy / m_size.cy) * m_size.cy;
	for(i = 0;i < cx / m_size.cx;i ++)
		m_pPicture->Render(dc, x + i * m_size.cx, y + Shift_Y, m_size.cx, cy - Shift_Y, 0, hmHeight, hmWidth, -hmHeight * (cy - Shift_Y)/m_size.cy, NULL);
	
	// Step4: Filling the right bottom corner
	m_pPicture->Render(dc, x + Shift_X, y + Shift_Y, cx - Shift_X, cy - Shift_Y, 0, hmHeight, hmWidth * (cx - Shift_X)/m_size.cx, -hmHeight * (cy - Shift_Y)/m_size.cy, NULL);
}

