/*
 * VXPLib COMPONENT
 *
 * Module: TOFont.h - TOFont.cpp - Implementations of QHTML tag classes
 * to work with fonts;
 *
 * Synopsis: This file contains implementations of all classes
 * that work with font-related tag objects of QHTML syntax.
 *
 * Copyright (c) 2002, Vitaly Tomilov
 * Copyright (c) 2003-2005, Tooltips.Net L.L.C.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * Library support web site - http://www.tooltips.net
 */

#include "stdafx.h"
#include "TOFont.h"

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

CTOFont::CTOFont(CTag * tag, CGlobalObject * globalObject):CTagObject(tag)
{
	m_pGlobalObject = globalObject;
	
	m_bHasCharSet = false;
	RetrieveCharSet(tag);

	bool IsWeightSpecified;
	int tmpWeight = tag->GetAttrIntValue(_T("WEIGHT"), &IsWeightSpecified);
	m_uWeight = IsWeightSpecified?tmpWeight:-1;

	bool IsSizeSpecified;
	int tmpSize = tag->GetAttrIntSignedValue(attrSIZE, m_SizeSign, &IsSizeSpecified);
	m_uSize = IsSizeSpecified?tmpSize:-1;

	bool IsFaceSpecified;
	m_strFace = tag->GetAttrTextValue(_T("FACE"), &IsFaceSpecified);
	if(IsFaceSpecified)
	{
		m_strFace.TrimLeft();
		m_strFace.TrimRight();
		m_strFace.Remove('\n');
		m_strFace.Remove('\r');
	}

	m_TextColor = tag->GetAttrColorValue(attrCOLOR, m_bColorValid);
	m_TextBkColor = tag->GetAttrColorValue(attrBGCOLOR, m_bBGColorValid);
}

// We only change the current char set, if the char set has been
// redifined, and it has been redefined correctly (it has been recognized).
void CTOFont::RetrieveCharSet(CTag * tag)
{
	CString char_set(tag->GetAttrTextValue(_T("CHARSET"), &m_bHasCharSet));
	if(!m_bHasCharSet)
		return;

	char_set.Remove(' ');
	if(char_set.IsEmpty())
	{
		m_bHasCharSet = false;
		return;
	}

	char_set.MakeUpper();

struct CCharSet
{
	LPCTSTR m_strName;
	BYTE m_Type;
};

#define CHARSET_NUMBER	19	// The number of supported charsets

CCharSet charsets[CHARSET_NUMBER] =
			{{_T("ANSI"), csANSI}, {_T("ARABIC"), csARABIC}, {_T("BALTIC"), csBALTIC}, {_T("CHINESEBIG5"), csCHINESEBIG5},
			{_T("DEFAULT"), csDEFAULT}, {_T("EASTEUROPE"), csEASTEUROPE}, {_T("GB2312"), csGB2312},
			{_T("GREEK"), csGREEK}, {_T("HANGUL"), csHANGUL}, {_T("HEBREW"), csHEBREW}, {_T("JOHAB"), csJOHAB},
			{_T("MAC"), csMAC}, {_T("OEM"), csOEM}, {_T("RUSSIAN"), csRUSSIAN}, {_T("SHIFTJIS"), csSHIFTJIS},
			{_T("SYMBOL"), csSYMBOL}, {_T("THAI"), csTHAI}, {_T("TURKISH"), csTURKISH}, {_T("VIETNAMESE"), csVIETNAMESE}};

	for(int i = 0;i < CHARSET_NUMBER;i ++)
		if(char_set == charsets[i].m_strName)
		{
			m_CharSet = charsets[i].m_Type;
			m_bHasCharSet = true;
			break;
		}
	if(!m_bHasCharSet)
	{
		m_CharSet = (BYTE)_ttol(char_set);
		m_bHasCharSet = true;
	}
}

LOGFONTEX CTOFont::SetupLogFont(HDC dc)
{
	LOGFONTEX LogFont;
	m_pGlobalObject->GetCurrentFont(&LogFont);
	if(m_uWeight != -1)
		LogFont.lfWeight = m_uWeight;
	if(m_strFace.GetLength())
		_copynstr(LogFont.lfFaceName, m_strFace, min(m_strFace.GetLength() + 1, LF_FACESIZE));

	if(m_uSize != -1)
	{
		LogFont.lfWidth = 0;
		if(m_SizeSign)
		{
			LOGFONT lf = m_pGlobalObject->GetDefaultLogFont();
			LogFont.lfHeight = lf.lfHeight - m_uSize * m_SizeSign;
		}
		else
			LogFont.lfHeight = -MulDiv(m_uSize, GetDeviceCaps(dc, LOGPIXELSY), 72);
	}
	
	if(m_bColorValid)
		LogFont.m_TextColor = m_TextColor;

	if(m_bBGColorValid)
	{
		LogFont.m_TextBkColor = m_TextBkColor;
		LogFont.m_bBkColorValid = true;
	}

	if(m_bHasCharSet)
		LogFont.lfCharSet = m_CharSet;

	return LogFont;
}

CSize CTOFont::GetSize(HDC dc)
{
	m_pGlobalObject->AddFont(&SetupLogFont(dc));
	return CSize();
}

void CTOFont::OnDrawTag(HDC dc, int x, int y, int width, int height, int ZeroX, int ZeroY)
{
	m_pGlobalObject->AddFont(&SetupLogFont(dc));
}

////////////////////////////////////////////////
// CTOFontEnd class implementation:
////////////////////////////////////////////////

CTOFontEnd::CTOFontEnd(CGlobalObject * globalObject):CTagObject(NULL)
{
	m_pGlobalObject = globalObject;
}

void CTOFontEnd::OnDrawTag(HDC dc, int x, int y, int width, int height, int ZeroX, int ZeroY)
{
	m_pGlobalObject->SetPreviousFont();
}

CSize CTOFontEnd::GetSize(HDC dc)
{
	m_pGlobalObject->SetPreviousFont();
	return CSize();
}

////////////////////////////////////////////////
// CTOBold class implementation:
////////////////////////////////////////////////

CTOBold::CTOBold(CTag * tag, CGlobalObject * globalObject):CTagObject(tag)
{
	m_pGlobalObject = globalObject;
}

void CTOBold::OnDrawTag(HDC dc, int x, int y, int width, int height, int ZeroX, int ZeroY)
{
	m_pGlobalObject->IncrementBold();
}

CSize CTOBold::GetSize(HDC dc)
{
	m_pGlobalObject->IncrementBold();
	return CSize();
}

////////////////////////////////////////////////
// CTOStrike class implementation:
////////////////////////////////////////////////

CTOStrike::CTOStrike(CTag * tag, CGlobalObject * globalObject):CTagObject(tag)
{
	m_pGlobalObject = globalObject;
}

void CTOStrike::OnDrawTag(HDC dc, int x, int y, int width, int height, int ZeroX, int ZeroY)
{
	m_pGlobalObject->IncrementStrike();
}

CSize CTOStrike::GetSize(HDC dc)
{
	m_pGlobalObject->IncrementStrike();
	return CSize();
}

////////////////////////////////////////////////
// CTOItalic class implementation:
////////////////////////////////////////////////

CTOItalic::CTOItalic(CTag * tag, CGlobalObject * globalObject):CTagObject(tag)
{
	m_pGlobalObject = globalObject;
}

void CTOItalic::OnDrawTag(HDC dc, int x, int y, int width, int height, int ZeroX, int ZeroY)
{
	m_pGlobalObject->IncrementItalic();
}

CSize CTOItalic::GetSize(HDC dc)
{
	m_pGlobalObject->IncrementItalic();
	return CSize();
}

////////////////////////////////////////////////
// CTOUnderline class implementation:
////////////////////////////////////////////////

CTOUnderline::CTOUnderline(CTag * tag, CGlobalObject * globalObject):CTagObject(tag)
{
	m_pGlobalObject = globalObject;
}

void CTOUnderline::OnDrawTag(HDC dc, int x, int y, int width, int height, int ZeroX, int ZeroY)
{
	m_pGlobalObject->IncrementUnderline();
}

CSize CTOUnderline::GetSize(HDC dc)
{
	m_pGlobalObject->IncrementUnderline();
	return CSize();
}

////////////////////////////////////////////////
// CTOBoldEnd class implementation:
////////////////////////////////////////////////

CTOBoldEnd::CTOBoldEnd(CGlobalObject * globalObject):CTagObject(NULL)
{
	m_pGlobalObject = globalObject;
}

void CTOBoldEnd::OnDrawTag(HDC dc, int x, int y, int width, int height, int ZeroX, int ZeroY)
{
	m_pGlobalObject->DecrementBold();
}

CSize CTOBoldEnd::GetSize(HDC dc)
{
	m_pGlobalObject->DecrementBold();
	return CSize();
}

////////////////////////////////////////////////
// CTOStrikeEnd class implementation:
////////////////////////////////////////////////

CTOStrikeEnd::CTOStrikeEnd(CGlobalObject * globalObject):CTagObject(NULL)
{
	m_pGlobalObject = globalObject;
}

void CTOStrikeEnd::OnDrawTag(HDC dc, int x, int y, int width, int height, int ZeroX, int ZeroY)
{
	m_pGlobalObject->DecrementStrike();
}

CSize CTOStrikeEnd::GetSize(HDC dc)
{
	m_pGlobalObject->DecrementStrike();
	return CSize();
}

////////////////////////////////////////////////
// CTOItalicEnd class implementation:
////////////////////////////////////////////////

CTOItalicEnd::CTOItalicEnd(CGlobalObject * globalObject):CTagObject(NULL)
{
	m_pGlobalObject = globalObject;
}

void CTOItalicEnd::OnDrawTag(HDC dc, int x, int y, int width, int height, int ZeroX, int ZeroY)
{
	m_pGlobalObject->DecrementItalic();
}

CSize CTOItalicEnd::GetSize(HDC dc)
{
	m_pGlobalObject->DecrementItalic();
	return CSize();
}

////////////////////////////////////////////////
// CTOUnderlineEnd class implementation:
////////////////////////////////////////////////

CTOUnderlineEnd::CTOUnderlineEnd(CGlobalObject * globalObject):CTagObject(NULL)
{
	m_pGlobalObject = globalObject;
}

void CTOUnderlineEnd::OnDrawTag(HDC dc, int x, int y, int width, int height, int ZeroX, int ZeroY)
{
	m_pGlobalObject->DecrementUnderline();
}

CSize CTOUnderlineEnd::GetSize(HDC dc)
{
	m_pGlobalObject->DecrementUnderline();
	return CSize();
}

