/*
 * VXPLib COMPONENT
 *
 * Module: TableMatrix.cpp - Implementation of classes CMatrixCell
 * and CTableMatrix;
 *
 * Synopsis: The QHTML parser implements matrix analysis of tag <TABLE>
 * Two classes CMatrixCell and CTableMatrix implement logic of syntax analysis
 * for tag <TABLE> of one level of depth.
 *
 * Copyright (c) 2002, Vitaly Tomilov
 * Copyright (c) 2003-2005, Tooltips.Net L.L.C.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * Library support web site - http://www.tooltips.net
 */

#include "stdafx.h"
#include "TableMatrix.h"
#include "TagLib.h"

CMatrixCell::CMatrixCell()
{
	m_pFirst = NULL;
	m_pLast = NULL;
	m_pTD = NULL;
	m_pTR = NULL;
}

void CMatrixCell::Init(CTagObject * first, CTagObject * last)
{
	m_pFirst = first;
	m_pLast = last;
}

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

CTableMatrix::CTableMatrix(CTagObject * start, CTagObject * end)
{
	m_pStart = start;
	m_pEnd = end;

	m_pCells = NULL;
	InitMatrix();
}

CTableMatrix::~CTableMatrix()
{
	ReleaseMatrix();
}

// This function must return the matrix size
// cx = cols
// cy = rows
CSize CTableMatrix::GetMatrixSize()
{
	return m_size;
}

// This method creates the matrix.
// In order for a line of values to be recognized az a row/column
// there must be both <TR> and <TD> tags. <TR> - first, <TD> - follows.
// They must be closed before the expression is recognized as row/column.
void CTableMatrix::InitMatrix()
{
	ReleaseMatrix();

// Step1: Calculate the matrix size
// The start object is <table>, the end object is </table>
// Between them can be another table and tags <TR> and <TD>,
// which shouldn't be counted.

	CSize size = CalculateSize();

// Step2: Allocate memory for the matrix
// In order to avoid further usage of whatever objects might reside
// inside the zero size table, they all must be set as unused.
	if(size.IsEmpty())
	{
		CTagObject * object = m_pStart;
		while(object != m_pEnd)
		{
			object->SetUnused();
			object = (CTagObject*)object->Next;
		}
		return;
	}
	
	m_size = size;

	m_pCells = new CMatrixCell**[m_size.cy];
	int i, k;

	for(i = 0;i < m_size.cy;i ++)
		m_pCells[i] = new CMatrixCell*[m_size.cx];

	for(i = 0;i < m_size.cy;i ++)
		for(k = 0;k < m_size.cx;k ++)
			m_pCells[i][k] = new CMatrixCell;

// Step3: Populate the matrix with values

// All elements not within correct <TR><TD>...</TD></TR>
// are to be ignored and marked UnUsed();
	PopulateMatrix();
}

void CTableMatrix::ReleaseMatrix()
{
	if(!m_pCells)
		return;

	for(int i = 0;i < m_size.cy;i ++)
	{
		for(int k = 0;k < m_size.cx;k ++)
			delete m_pCells[i][k];
		delete m_pCells[i];
	}
	delete m_pCells;

	m_pCells = NULL;
	m_size.Zero();
}

CMatrixCell * CTableMatrix::GetCell(int col, int row)
{
	return m_pCells[row][col];
}

// This function must return a size of the table in cols/rows.
// start = <table>; end = </table>
// tags <tr> and <tr>, if inside another table must
// start->Next != end
// be ignored.
// Important: Should be a cell generated, when it's empty?
// Yes, it is generated anyway!
// There can be another <table></table>, but there cannot be
// </table> without <table>
// Rows without columns at all do not count
// Tags <TR> and <TD> cannot be inside of each other (recursive).
// If <TR> is inside of another <TR> or <TD>, or <TD> is inside
// of another <TD>, such object is marked as unused:
// object->SetUnused();
// All objects inside wrong table context are to be marked as unused
CSize CTableMatrix::CalculateSize()
{
	CTagObject * object = (CTagObject*)m_pStart->Next;

	m_MaxColsRow = 0;

	// These two are just boolean, because <TR> and <TD> do
	// not support recursion.
	bool TR = false;	// Indicates true, if we are inside TR
	bool TD = false;	// Indicates true, if we are inside TD
	
	bool td_closed = true; // Indicates true, if TD reverted at least once

	int table = 0;	// Counter of tables inside
	
	int rows = 0;	// Counter of rows
	int cols = 0;	// Counter of columns
	int max_cols = 0; // Maximum cols

	while(object != m_pEnd)
	{
		switch(object->GetTagAttr())
		{
		case TA_TABLE:
			{
				// Increment tables counter to avoid wrong counting for cols/rows
				table ++; 
				break;
			};
		case TA_TABLE_END:
			{
				table --; // It cannot get less than 0. no need to verify
				break;
			};
		case TA_TR:
			{
				if(!table && !TR && !TD)
					TR = true;
				break;
			};
		case TA_TR_END:	// Only this tag first met can say there's something
			{
				if(!table && TR)
				{
					rows ++;
					TR = false;
					TD = false;
					td_closed = true;
					if(cols > max_cols)
					{
						m_MaxColsRow = rows - 1;
						max_cols = cols;
					}
					cols = 0;
				}
				break;
			};
		case TA_TD:
			{
				if(!table && TR && !TD && td_closed) // Without <TR>, <TD> is ignored
				{
					td_closed = false;
					TD = true;
				}
				break;
			};
		case TA_TD_END: // We assume this row will close
			{
				if(!table && TR && TD && !td_closed)
				{
					TD = false;
					td_closed = true;
					cols ++;
				}
				break;
			};
		default:
			break;
		};
		object = (CTagObject*)object->Next;
	}
	return CSize(max_cols, rows);
}

// All elements not within correct <TR><TD>...</TD></TR>
// are to be ignored and marked UnUsed();
void CTableMatrix::PopulateMatrix()
{
	CTagObject * object = (CTagObject*)m_pStart->Next;

	// These two are just boolean, because <TR> and <TD> do
	// not support recursion.
	bool TR = false;	// Indicates true, if we are inside TR
	bool TD = false;	// Indicates true, if we are inside TD
	
	bool td_closed = true; // Indicates true, if TD reverted at least once

	int table = 0;	// Counter of tables inside
	
	int rows = 0;	// Counter of rows
	int cols = 0;	// Counter of columns
	int max_cols = 0; // Maximum cols

	CTagObject * pLastTD;

	CTagObject * first = NULL;

	while(object != m_pEnd)
	{
		switch(object->GetTagAttr())
		{
		case TA_TABLE:
			{
				// Increment tables counter to avoid wrong counting for cols/rows
				table ++; 
				break;
			};
		case TA_TABLE_END:
			{
				table --; // It cannot get less than 0. no need to verify
				break;
			};
		case TA_TR:
			{
				if(!table && !TR && !TD)
				{
					TR = true;
					first = object;
				}
				break;
			};
		case TA_TR_END:	// Only this tag first met can say there's something
			{
				if(!table && TR)
				{
					//*********
					PopulateRow(first, object, cols, rows);
					first = NULL;
					//*********
					TD = false;
					if(!td_closed) // Means  TD wasn't closed
					{
						td_closed = true;
						// Setting all tags in between to unused:
						CTagObject * pObject = pLastTD;
						while(pObject != object)
						{
							pObject->SetUnused();
							pObject = (CTagObject*)pObject->Next;
						}
					}
					rows ++;
					TR = false;
					if(cols > max_cols)
						max_cols = cols;
					cols = 0;
				}
				break;
			};
		case TA_TD:
			{
				if(!table && TR && !TD && td_closed) // Without <TR>, <TD> is ignored
				{
					pLastTD = object;
					td_closed = false;
					TD = true;
				}
				break;
			};
		case TA_TD_END: // We assume this row will close
			{
				if(!table && TR && TD && !td_closed)
				{
					TD = false;
					td_closed = true;
					cols ++;
				}
				break;
			};
		default:
			{
				if(!table && !TD)
					object->SetUnused();
				break;
			}
		};
		object = (CTagObject*)object->Next;
	}
}

// start = <TR>, end = </TR>,
// just <TD>, </TD> and other objects
// There also can be <table>, which is always closed by </table>
// Unused elements are to be marked as such
// nCols = the number of columns in this row
// row = the row index
void CTableMatrix::PopulateRow(CTagObject *start, CTagObject *end, int nCols, int row)
{
	CTagObject * object = (CTagObject*)start->Next;

	for(int i = 0;i < nCols;i ++)
	{
		int skip = 0;
		int attr;
		while((attr = object->GetTagAttr()) != TA_TD || skip)
		{
			switch(attr)
			{
			case TA_TABLE:
				{
					skip ++;
					break;
				}
			case TA_TABLE_END:
				{
					skip --;
					break;
				}
			default:
				{
					if(!skip) // If not with an inside table
						object->SetUnused();
					break;
				}
			}
			object = (CTagObject*)object->Next;
		}
		CTagObject * begin = (CTagObject*)object->Next;
		CTagObject * TD = object;
		object = begin;
		while((attr = object->GetTagAttr()) != TA_TD_END || skip)
		{
			switch(attr)
			{
			case TA_TABLE:
				{
					skip ++;
					break;
				}
			case TA_TABLE_END:
				{
					skip --;
					break;
				}
			case TA_TD:
			case TA_TR:
			case TA_TR_END:
				{
					if(!skip)
						object->SetUnused();
					break;
				}
			default:
				{

					break;
				}
			}
			object = (CTagObject*)object->Next;
		}
		
		m_pCells[row][i]->m_pTR = start;	// Store initial TR in all cells
		m_pCells[row][i]->m_pTD = TD;		// Store initial TD in all cells

		m_pCells[row][i]->m_MinSize.SetSize(TD->GetForcedSize());
		if(begin != object)
			m_pCells[row][i]->Init(begin, (CTagObject*)object->Prev);
	}
}

bool CMatrixCell::IsEmpty()
{
	return (!m_pFirst || !m_pLast);
}

// This function must set all coordinate parameters
// for all the table elements (TR and TD).
// width = calculated table width in pixels, including borders
// height = calculated table height in pixels, including borders
// border = table border size
// Extra parameters to be used: m_cellSize and m_contentSize
void CTableMatrix::SetBorderPos(int ZeroX, int ZeroY, int width, int height, int border, int cellspacing)
{
	// Setting the table coordinates:
	m_pStart->SetClientRect(ZeroX, ZeroY, width, height);

	CTagObject * object = (CTagObject*)m_pStart->Next;

	// These two are just boolean, because <TR> and <TD> do
	// not support recursion.
	bool TR = false;	// Indicates true, if we are inside TR
	bool TD = false;	// Indicates true, if we are inside TD
	
	bool td_closed = true; // Indicates true, if TD reverted at least once

	int table = 0;	// Counter of tables inside
	
	int rows = 0;	// Counter of rows
	int cols = 0;	// Counter of columns
	int max_cols = 0; // Maximum cols

	CTagObject * LastTR;
	int cell_height = 0;
	int cell_width = 0;
	int col_index = 0;

	CSize matr_size = GetMatrixSize();

	while(object != m_pEnd)
	{
		switch(object->GetTagAttr())
		{
		case TA_TABLE:
			{
				// Increment tables counter to avoid wrong counting for cols/rows
				table ++; 
				break;
			};
		case TA_TABLE_END:
			{
				table --; // It cannot get less than 0. no need to verify
				break;
			};
		case TA_TR:
			{
				if(!table && !TR && !TD)
				{
					TR = true;
					LastTR = object;
				}
				break;
			};
		case TA_TR_END:	// Only this tag first met can say there's something
			{
				if(!table && TR)
				{
					LastTR->m_pTable = m_pStart;
					rows ++;
					TR = false;
					TD = false;
					td_closed = true;
					if(cols > max_cols)
						max_cols = cols;
					cols = 0;
					//*********
					if(rows <= matr_size.cy)
					{
						cell_height += GetCell(0, rows - 1)->m_cellSize.cy;
						LastTR->SetClientRect(ZeroX + border, ZeroY + cell_height + cellspacing * (rows-1)+border, width - border * 2, cellspacing);

						if(rows == matr_size.cy)
							((CTOTableTR*)LastTR)->SetLast();	// This is to prevent drawing a horizintal line
												// for the very last row of the table
					}
				}
				break;
			};
		case TA_TD:
			{
				if(!table && TR && !TD && td_closed && !object->IsUnused())
				{
					td_closed = false;
					TD = true;
					object->m_pTable = m_pStart;
					if(rows == m_MaxColsRow && col_index + 1 < matr_size.cx)
					{
						cell_width += GetCell(col_index, m_MaxColsRow)->m_cellSize.cx;
						col_index ++;
						object->SetClientRect(ZeroX + border + cellspacing * (col_index - 1)+ cell_width, ZeroY + border, cellspacing, height - border * 2);
					}
				}
				break;
			};
		case TA_TD_END: // We assume this row will close
			{
				if(!table && TR && TD && !td_closed)
				{
					TD = false;
					td_closed = true;
					cols ++;
				}
				break;
			};
		default:
			break;
		};
		object = (CTagObject*)object->Next;
	}
}

int CTableMatrix::GetMaxColsRow()
{
	return m_MaxColsRow;
}
