/*
 * VXPLib COMPONENT
 *
 * Module: TagLib.cpp - Implementation of QHTML tag classes
 *
 * Synopsis: This file contains implementations of all classes
 * that represent tag objects of QHTML syntax.
 *
 * Copyright (c) 2002, Vitaly Tomilov
 * Copyright (c) 2003-2005, Tooltips.Net L.L.C.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * Library support web site - http://www.tooltips.net
 */

#include "stdafx.h"
#include "TagLib.h"

// *************************************************
// CTOTable implementation *************************
// *************************************************

CTOTable::CTOTable(CTag * tag, CGlobalObject * globalObject):CTagObject(tag)
{
	SetTagAttr(TA_TABLE);
	m_uBorder = tag->GetAttrIntValue(attrBORDER);

// Gradient support: *********************************
	m_grad.Init(tag);
// ***************************************************

	bool found;
	m_uCellSpacing = tag->GetAttrIntValue(_T("CS"), &found);
	if(!found)
		m_uCellSpacing = tag->GetAttrIntValue(_T("CELLSPACING"), &found);

	m_uCellPadding = 2;
	int cellpadding = tag->GetAttrIntValue(_T("CP"), &found);
	if(!found)
		cellpadding = tag->GetAttrIntValue(_T("CELLPADDING"), &found);
	if(found)
		m_uCellPadding = cellpadding;

	m_bgColor = tag->GetAttrColorValue(attrBGCOLOR, m_bIsBGValid);

	m_BorderColor = tag->GetAttrColorValue(attrBC, m_bIsBCValid);
	if(!m_bIsBCValid)
		m_BorderColor = tag->GetAttrColorValue(attrBORDERCOLOR, m_bIsBCValid);

	m_pBackground = CPicture::GetObjectFromTag(tag, NULL, globalObject);
}

CTOTable::~CTOTable()
{
	if(m_pBackground)
		delete m_pBackground;
}

COLORREF CTOTable::GetBorderColor()
{
	return m_bIsBCValid?m_BorderColor:RGB(0,0,0);
}

int CTOTable::GetBorderSize()
{
	return m_uBorder;
}

void CTOTable::OnDrawTag(HDC dc, int x, int y, int width, int height, int ZeroX, int ZeroY)
{
	if(m_grad.m_bGFCValid)
		FillGradient(dc, m_bgColor, m_grad.m_GFC, x, y, x + width, y + height, m_grad.m_bGFCHorizontal);
	else
		if(m_bIsBGValid)
		{
			HBRUSH bg = CreateSolidBrush(m_bgColor);
			RECT r = {x, y, x + width, y + height};
			FillRect(dc, &r, bg);
			DeleteObject(bg);
		}

	RECT r1 = {x, y, x + m_uBorder, y + height};
	RECT r2 = {x + width - m_uBorder, y, x + width, y + height};
	RECT r3 = {x, y, x + width, y + m_uBorder};
	RECT r4 = {x, y + height - m_uBorder, x + width, y + height};

	HBRUSH brush = CreateSolidBrush(GetBorderColor());

	FillRect(dc, &r1, brush);
	FillRect(dc, &r2, brush);
	FillRect(dc, &r3, brush);
	FillRect(dc, &r4, brush);
	DeleteObject(brush);

	if(m_pBackground)
		m_pBackground->Populate(dc, x + m_uBorder, y + m_uBorder, width - m_uBorder * 2, height - m_uBorder * 2);
}

int CTOTable::GetCellPadding()
{
	return m_uCellPadding;
}

int CTOTable::GetCellSpacing()
{
	return m_uCellSpacing;
}

// *************************************************
// CTOTableEnd implementation **********************
// *************************************************
CTOTableEnd::CTOTableEnd(CTag * tag):CTagObject(tag)
{
	SetTagAttr(TA_TABLE_END);
}

// *************************************************
// CTOTableTR implementation ***********************
// *************************************************
CTOTableTR::CTOTableTR(CTag * tag, CGlobalObject * globalObject):CTagObject(tag)
{
	m_bLast = false;
	SetTagAttr(TA_TR);
	SetHAlignment(TextIntoAlignment(tag->GetAttrTextValue(attrALIGN)));
	SetVAlignment(TextIntoAlignment(tag->GetAttrTextValue(attrVALIGN)));
	m_bgColor = tag->GetAttrColorValue(attrBGCOLOR, m_bIsBGValid);
	m_pBackground = CPicture::GetObjectFromTag(tag, NULL, globalObject);

// Gradient support: *********************************
	m_grad.Init(tag);
// ***************************************************
}

CTOTableTR::~CTOTableTR()
{
	if(m_pBackground)
		delete m_pBackground;
}

void CTOTableTR::OnDrawTag(HDC dc, int x, int y, int width, int height, int ZeroX, int ZeroY)
{
	if(!m_pTable)
		return;

	RECT r = {ZeroX + m_FullRect.left, ZeroY + m_FullRect.top, ZeroX + m_FullRect.left + m_FullRect.right, ZeroY + m_FullRect.top + m_FullRect.bottom};

	if(m_grad.m_bGFCValid)
		FillGradient(dc, m_bgColor, m_grad.m_GFC, r.left, r.top, r.right, r.bottom, m_grad.m_bGFCHorizontal);
	else
		if(m_bIsBGValid)
		{
			HBRUSH bg = CreateSolidBrush(m_bgColor);
			FillRect(dc, &r, bg);
			DeleteObject(bg);
		}

	if(!m_bLast)
	{
		RECT r2 = {x, y, x + width, y + height};
		HBRUSH brush = CreateSolidBrush(((CTOTable*)m_pTable)->GetBorderColor());
		FillRect(dc, &r2, brush);
		DeleteObject(brush);
	}

	if(m_pBackground)
		m_pBackground->Populate(dc, ZeroX + m_FullRect.left, ZeroY + m_FullRect.top, m_FullRect.right, m_FullRect.bottom);
}

void CTOTableTR::SetLast()
{
	m_bLast = true; // The row is the very last one in the table
}

// *************************************************
// CTOTableTREnd implementation ********************
// *************************************************
CTOTableTREnd::CTOTableTREnd(CTag * tag):CTagObject(tag)
{
	SetTagAttr(TA_TR_END);
}

// *************************************************
// CTOTableTD implementation ***********************
// *************************************************
CTOTableTD::CTOTableTD(CTag * tag, CGlobalObject * globalObject):CTagObject(tag)
{
	globalObject->m_bNewLine = true;

	SetTagAttr(TA_TD);
	SetHAlignment(TextIntoAlignment(tag->GetAttrTextValue(attrALIGN)));
	SetVAlignment(TextIntoAlignment(tag->GetAttrTextValue(attrVALIGN)));
	m_MinSize.cx = tag->GetAttrIntValue(attrWIDTH);
	m_MinSize.cy = tag->GetAttrIntValue(attrHEIGHT);
	m_bgColor = tag->GetAttrColorValue(attrBGCOLOR, m_bIsBGValid);
	m_pBackground = CPicture::GetObjectFromTag(tag, NULL, globalObject);

// Gradient support: *********************************
	m_grad.Init(tag);
// ***************************************************
}

CTOTableTD::~CTOTableTD()
{
	if(m_pBackground)
		delete m_pBackground;
}

void CTOTableTD::OnDrawTag(HDC dc, int x, int y, int width, int height, int ZeroX, int ZeroY)
{
	if(!m_pTable)
		return;

	// This is a normal code to draw the inside color of TD
	RECT r = {ZeroX + m_FullRect.left, ZeroY + m_FullRect.top, ZeroX + m_FullRect.left + m_FullRect.right, ZeroY + m_FullRect.top + m_FullRect.bottom};
	if(m_grad.m_bGFCValid)
		FillGradient(dc, m_bgColor, m_grad.m_GFC, r.left, r.top, r.right, r.bottom, m_grad.m_bGFCHorizontal);
	else
		if(m_bIsBGValid)
		{		
			HBRUSH bg = CreateSolidBrush(m_bgColor);
			FillRect(dc, &r, bg);
			DeleteObject(bg);
		}

	if(m_pBackground)
		m_pBackground->Populate(dc, ZeroX + m_FullRect.left, ZeroY + m_FullRect.top, m_FullRect.right, m_FullRect.bottom);
}

void CTOTableTD::OnDrawTDBorders(HDC dc, int ZeroX, int ZeroY)
{
	if(m_pTable)	// This is to fix a bug for now. This object shouldn't exist
					// with having its m_pTable set to NULL. Needs to be fixed later,
					// and remove the verification. Fails on huge HTML files.
					// P.S. Most likely fixed (indirectly), but left the verification
					// just in case, as testing is very complex.
	{
		RECT r = {ZeroX + m_client.left, ZeroY + m_client.top, ZeroX + m_client.left + m_client.right, ZeroY + m_client.top + m_client.bottom};
		HBRUSH brush = CreateSolidBrush(((CTOTable*)m_pTable)->GetBorderColor());
		FillRect(dc, &r, brush);
		DeleteObject(brush);
	}
}

CSize CTOTableTD::GetForcedSize()
{
	return m_MinSize;
}

// *************************************************
// CTOTableTDEnd implementation ********************
// *************************************************
CTOTableTDEnd::CTOTableTDEnd(CTag * tag):CTagObject(tag)
{
	SetTagAttr(TA_TD_END);
}

// **************************************************
// CTOComment implementation ************************
// **************************************************
CTOComment::CTOComment(CString text):CTagObject(NULL)
{
	SetTagAttr(TA_COMMENT);
	m_text = text;
	
	m_text.TrimLeft();
	m_text.TrimRight();
	if(m_text.GetLength())
		m_text += _T(" ");
}

CString CTOComment::GetText()
{
	return m_text;
}

// *************************************************
// CTOSound implementation *************************
// *************************************************
CTOSound::CTOSound(CString text):CTagObject(NULL)
{
	SetTagAttr(TA_SOUND);
	m_text = text;
}

CString CTOSound::GetText()
{
	return m_text;
}

// *************************************************
// CTOText implementation **************************
// *************************************************
CTOText::CTOText(CString text, CGlobalObject * globalObject, bool bCutSpaces):CTagObject(NULL)
{
	m_pGlobalObject = globalObject;

	if(globalObject->m_bNewLine) // If this is the first text line
	{
		text.TrimLeft(); // Remove all preceeding spaces
		if(!text.IsEmpty())
			globalObject->m_bNewLine = false;
	}

	m_tmDescent = 0;
	text.SmartTrim(bCutSpaces);
	if(!text.IsEmpty())
		m_text = CTag::SetSpecialSymbols(text);
}

CString CTOText::GetText()
{
	return m_text;
}

CSize CTOText::GetSize(HDC dc)
{
	if(m_text.IsEmpty())
		return CSize(0, 0);
	// Return the text size here
	HGDIOBJ OldObject = m_pGlobalObject->SelectCurrentFont(dc);
	CSize s = m_pGlobalObject->GetTextSize(dc, m_text);
	if(m_pGlobalObject->IsItalic())
	{
		TEXTMETRIC tm;
		GetTextMetrics(dc, &tm);

// How to make it look perfectly correct with italic fonts is still
// an open question;
// An average gap is impossible...??? Then what calculation can be done?
//		1) s.cx += tm.tmAveCharWidth / 8; ? - sometimes too small gap
//		2) s.cx += tm.tmOverhang; ? - sometimes too large gap

	}
	SelectObject(dc, OldObject);
	return s;
}

int CTOText::GetTopHeight(HDC dc)
{
	if(m_text.IsEmpty())
		return 0;
	HGDIOBJ OldObject = m_pGlobalObject->SelectCurrentFont(dc);
	TEXTMETRIC tm;
	GetTextMetrics(dc, &tm);
	SelectObject(dc, OldObject);
	m_tmDescent = tm.tmDescent;
	m_AveWidth = tm.tmAveCharWidth;
	return tm.tmAscent;
}

void CTOText::OnDrawTag(HDC dc, int x, int y, int width, int height, int ZeroX, int ZeroY)
{
	if(m_text.IsEmpty())
		return;
	HGDIOBJ OldObject = m_pGlobalObject->SelectCurrentFont(dc);
	if(m_MaxLine)
		y += m_MaxLine + m_tmDescent - height;
	m_pGlobalObject->TextOut(dc, x, y, m_text);
	SelectObject(dc, OldObject);
}

// *************************************************
// CTO_HR implementation ***************************
// *************************************************
CTO_HR::CTO_HR(CTag * tag, CGlobalObject * globalObject):CTagObject(tag)
{
	globalObject->m_bNewLine = true;

	SetTagAttr(TA_HR);
	SetNextLineJump(1); // Jump by two lines
	m_size = 2;
	int s = tag->GetAttrIntValue(attrSIZE);
	if(s < 1000 && s > 0)
		m_size = s;

	m_align = TextIntoAlignment(tag->GetAttrTextValue(attrALIGN));
	m_color = RGB(0,0,0);
	bool isSpecified;
	COLORREF tmpColor = tag->GetAttrColorValue(attrCOLOR, isSpecified);
	if(isSpecified)
		m_color = tmpColor;

	m_width = 100; // In percent
	
	short widthTmp = tag->GetAttrPercentValue(attrWIDTH, isSpecified);
	if(isSpecified)
		m_width = widthTmp;

	m_shape = HRS_RECT;
	CString tmpShape = tag->GetAttrTextValue(_T("SHAPE"));
	tmpShape.Remove(' ');
	tmpShape.MakeUpper();
	if(!tmpShape.IsEmpty())
	{
		if(tmpShape == _T("ROUND"))
			m_shape = HRS_ROUND;
		else
			if(tmpShape == _T("ELLIPSE"))
				m_shape = HRS_ELLIPSE;
	}
}

#define HR_MARGIN	4	// Margin for HR above and under

CSize CTO_HR::GetSize(HDC)
{
	// HR has cx normally zero, and cy depends on
	// CY = SIZE + HR_MARGIN * 2, to make sure
	// the object doesn't merge with others
	return CSize(0, m_size + HR_MARGIN * 2);
}

void CTO_HR::OnDrawTag(HDC dc, int x, int y, int width, int height, int ZeroX, int ZeroY)
{
	if(m_align == ALIGN_RIGHT)
		x += width - width*m_width/100;
	else
		if(m_align == ALIGN_CENTER)
			x += (width - width * m_width / 100) / 2;

	RECT r = {x, y + HR_MARGIN, x + width*m_width/100, y + height - HR_MARGIN};
	HBRUSH brush = CreateSolidBrush(m_color);
	HGDIOBJ old = SelectObject(dc, brush);
	switch(m_shape)
	{
	case HRS_RECT:
		{
			FillRect(dc, &r, brush);
			break;
		}
	case HRS_ROUND:
		{
			HPEN pen = CreatePen(PS_SOLID, 1, m_color);
			HGDIOBJ oldPen = SelectObject(dc, pen);
			int round = 2;
			round += height/8;
			RoundRect(dc, r.left, r.top, r.right, r.bottom, round, round);
			SelectObject(dc, oldPen);
			DeleteObject(pen);
			break;
		}
	case HRS_ELLIPSE:
		{
			HPEN pen = CreatePen(PS_SOLID, 1, m_color);
			HGDIOBJ oldPen = SelectObject(dc, pen);
			Ellipse(dc, r.left, r.top, r.right, r.bottom);
			SelectObject(dc, oldPen);
			DeleteObject(pen);
			break;
		}
	default:
		break;
	}
	SelectObject(dc, old);
	DeleteObject(brush);
}

// *************************************************
// CTO_BR implementation ***************************
// *************************************************
CTO_BR::CTO_BR(CTag * tag, CGlobalObject * globalObject):CTagObject(tag)
{
	globalObject->m_bNewLine = true;

	SetTagAttr(TA_BR);
	SetNextLineJump(1);	// Jump by one text line
	m_pGlobalObject = globalObject;
}

CSize CTO_BR::GetSize(HDC dc)
{
	HGDIOBJ OldObject = m_pGlobalObject->SelectCurrentFont(dc);
	TEXTMETRIC tm;
	GetTextMetrics(dc, &tm);
	SelectObject(dc, OldObject);
	return CSize(0, tm.tmHeight);
}

// *************************************************
// CTOImage implementation *************************
// *************************************************

CTOImage::CTOImage(CTag * tag, CGlobalObject * globalObject):CTagObject(tag)
{
// Gradient support: *********************************
	m_grad.Init(tag);
// ***************************************************

	m_alt = CTag::SetSpecialSymbols(tag->GetAttrTextValue(_T("ALT")));
	m_pGlobalObject = globalObject;
	m_border = tag->GetAttrIntValue(attrBORDER);
	if(m_border > 100)
		m_border=100;
	if(m_border < 0)
		m_border = 0;

	bool IsValueSpecified;
	int tmpValue = tag->GetAttrIntValue(attrWIDTH, &IsValueSpecified);
	m_Width = IsValueSpecified?tmpValue:-1;

	tmpValue = tag->GetAttrIntValue(attrHEIGHT, &IsValueSpecified);
	m_Height = IsValueSpecified?tmpValue:-1;

	// If at least one of them is zero, than the other must be zero too.
	if(!m_Height || !m_Width)
	{
		m_pPicture = NULL;
		m_Width = 0;
		m_Height = 0;
	}
	else
	{
		m_pPicture = CPicture::GetObjectFromTag(tag, _T("SRC"), globalObject, (m_Width>0)?m_Width:0, (m_Height>0)?m_Height:0);
		if(m_pPicture)
		{
			CSize s = m_pPicture->GetSize();
			if(m_Width < 0)
				m_Width = s.cx;
			if(m_Height < 0)
				m_Height = s.cy;
		}
		else
		{
			if(m_alt.IsEmpty())
			{
				m_Width = 0;
				m_Height = 0;
			}
			else
			{
				if(m_Width < 0)
					m_Width = 50;
				if(m_Height < 0)
					m_Height = 50;
			}
		}
	}
	m_bgColor = tag->GetAttrColorValue(attrBGCOLOR, m_BGColorValid);
	m_BorderColor = tag->GetAttrColorValue(attrBC, m_BColorValid);
	if(!m_BColorValid)
		m_BorderColor = tag->GetAttrColorValue(attrBORDERCOLOR, m_BColorValid);
}

CTOImage::~CTOImage()
{
	if(m_pPicture)
		delete m_pPicture;
}

CSize CTOImage::GetSize(HDC dc)
{
	return CSize(m_Width + m_border * 2, m_Height + m_border * 2);
}

// This margin is for the ALT text output
#define ALT_MARGIN 2

void CTOImage::OnDrawTag(HDC dc, int x, int y, int width, int height, int ZeroX, int ZeroY)
{
	RECT r = {x + m_border, y + m_border, x + width - m_border, y + height - m_border};
	if(m_grad.m_bGFCValid)
		FillGradient(dc, m_bgColor, m_grad.m_GFC, r.left, r.top, r.right, r.bottom, m_grad.m_bGFCHorizontal);
	else
		if(m_BGColorValid)
		{
			HBRUSH brush = CreateSolidBrush(m_bgColor);
			FillRect(dc, &r, brush);
			DeleteObject(brush);
		}
	// No need to verify that
	if(/*m_BColorValid && */m_border)
	{
		HBRUSH brush = CreateSolidBrush(m_BorderColor);
		RECT r = {x, y, x + m_border, y + height};
		FillRect(dc, &r, brush);
		RECT r2 = {x + m_border, y, x + width, y + m_border};
		FillRect(dc, &r2, brush);
		RECT r3 = {x + width - m_border, y + m_border, x + width, y + height};
		FillRect(dc, &r3, brush);
		RECT r4 = {x + m_border, y + height - m_border, x + width - m_border, y + height};
		FillRect(dc, &r4, brush);
		DeleteObject(brush);
	}
	if(m_pPicture)
		m_pPicture->Draw(dc, x + m_border, y + m_border, width - m_border, height - m_border);
	else
		if(!m_alt.IsEmpty())
		{
			RECT r = {x + m_border + ALT_MARGIN, y + m_border + ALT_MARGIN, x + width - m_border - ALT_MARGIN, y + height - m_border - ALT_MARGIN};
			HGDIOBJ OldObject = m_pGlobalObject->SelectCurrentFont(dc);
			DrawTextEx(dc, (LPTSTR)((LPCTSTR)m_alt), m_alt.GetLength(), &r, DT_CENTER|DT_EXPANDTABS|DT_VCENTER|DT_WORDBREAK, NULL);
			SelectObject(dc, OldObject);
		}
}

// *************************************************
// CTOTxt implementation ***************************
// *************************************************

#define MAX_FILE_SIZE 65536

CTOTxt::CTOTxt(CTag * tag, CGlobalObject * globalObject):CTagObject(NULL)
{
	m_pList = NULL;
	m_tmDescent = 0;
	m_pGlobalObject = globalObject;
	m_nSizeLimit = tag->GetAttrIntValue(attrSIZE);
//**********************************
	m_CodePage = m_pGlobalObject->GetTxtCodePage();
	bool found;
	int CodePage = tag->GetAttrIntValue(_T("CP"), &found);
	if(!found)
		CodePage = tag->GetAttrIntValue(_T("CODEPAGE"), &found);
	if(found)
	{
		m_CodePage = CodePage;
		if(!IsValidCodePage(m_CodePage))
		{
			m_CodePage = 0;
			m_pGlobalObject->AddError(EX_UNSUPPORTEDCODEPAGE);
		}
	}
//**********************************
	m_text = GetTextFromTag(tag);
	m_nFormat = TF_HTML;
	if(m_text.IsEmpty())
		return;
	m_nFormat = GetFormatType(tag);
	switch(m_nFormat)
	{
	case TF_SIMPLE:
		{
			FormatSimple();
			break;
		}
	case TF_HTML:
		{
			FormatHTML();
			break;
		}
	case TF_NONE:
		{
			FormatNone();
			break;
		}
	default:
		break;
	}
}

void CTOTxt::FormatNone()
{
	m_pList = new CTOText(m_text, m_pGlobalObject);
}

void CTOTxt::FormatHTML()
{
	m_text.Replace('\n', ' ');
	m_text.Replace('\r', ' ');
	m_text.Replace('\t', ' ');
	m_text.ShrinkEffective();
}

void CTOTxt::FormatSimple()
{
	m_text.Remove('\r');
	int L = m_text.GetLength();
	if(!L)
		return;
	int index = 0;
	int prev_index = 0;
	CTOText * txt;
	do
	{
		index = m_text.Find('\n', index);
		if(index == -1)
		{
			if(prev_index)
				txt = new CTOText(m_text.Right(L - prev_index), m_pGlobalObject);
			else
				txt = new CTOText(m_text, m_pGlobalObject);
			ADD_COLLECTION(m_pList, txt)
		}
		else
		{
			txt = new CTOText(m_text.Mid(prev_index, index - prev_index), m_pGlobalObject);
			ADD_COLLECTION(m_pList, txt)
			txt->Add(new CTO_BR(NULL, m_pGlobalObject));
			index ++;
			prev_index = index;
		}
	}
	while(index != -1 && index < L);
}

int CTOTxt::GetFormatType(CTag * tag)
{
	bool IsFormatSpecified;
	CString format = tag->GetAttrTextValue(_T("FORMAT"), &IsFormatSpecified);
	if(IsFormatSpecified)
	{
		format.Remove(' ');
		format.MakeUpper();
		if(format == _T("NONE"))
			return TF_NONE;
		if(format == _T("SIMPLE"))
			return TF_SIMPLE;
		if(format == _T("HTML"))
			return TF_HTML;
	}
	int N = m_text.Count('<');
	if(N > 0 && N == m_text.Count('>'))
		return TF_HTML;
	return TF_SIMPLE;
}

CSize CTOTxt::GetSize(HDC dc)
{
	if(m_text.IsEmpty())
		return CSize(0, 0);
	// Return the text size here
	HGDIOBJ OldObject = m_pGlobalObject->SelectCurrentFont(dc);
	CSize s = m_pGlobalObject->GetTextSize(dc, m_text);
	SelectObject(dc, OldObject);
	return s;
}

int CTOTxt::GetTopHeight(HDC dc)
{
	if(m_text.IsEmpty())
		return 0;
	HGDIOBJ OldObject = m_pGlobalObject->SelectCurrentFont(dc);
	TEXTMETRIC tm;
	GetTextMetrics(dc, &tm);
	SelectObject(dc, OldObject);
	m_tmDescent = tm.tmDescent;
	return tm.tmAscent;
}

void CTOTxt::OnDrawTag(HDC dc, int x, int y, int width, int height, int ZeroX, int ZeroY)
{
	if(m_text.IsEmpty())
		return;
	HGDIOBJ OldObject = m_pGlobalObject->SelectCurrentFont(dc);
	if(m_MaxLine)
		y += m_MaxLine + m_tmDescent - height;
	m_pGlobalObject->TextOut(dc, x, y, m_text);
	SelectObject(dc, OldObject);
}

CString CTOTxt::GetTextFromFile(CTag * tag)
{
	CString result;
	CString file = tag->GetAttrTextValue(attrFILE);
	file = m_pGlobalObject->GetValidFilePath(file);
	if(file.IsEmpty())
		return result;
//*** Reading the file *******
	int size;
	long limit = (m_nSizeLimit>0)?m_nSizeLimit:MAX_FILE_SIZE;
	BYTE * buffer = ReadFromFile(file, size, limit);
	if(size)
		result.TakeFromANSI((LPCSTR)buffer, size, m_CodePage);
	if(buffer)
		delete []buffer;
//****************************
	return result;
}

// RES Formats:
// 1: 'custom=#234324', or 'custom:#54566'
// 2: '453545' - for a simple text from the resource file
// If defined as 'html=...' then, loads from HTML instead.
// The second parameter 'res' is taken as reference only for speed
// optimization.
CString CTOTxt::GetTextFromRes(HINSTANCE hInst, CString & res)
{
	CString result;
	res.Remove(' ');
	if(res.IsEmpty())
		return result;

	int index = res.Find('=');
	if(index == -1)
		index = res.Find(':');
	if(index == -1)	// Simple text resource
	{
		res.Remove('#');
		int resID = _ttol(res);
		if(!resID)
			return result;
		result.LoadString(hInst, resID, (WORD)m_nSizeLimit);
		result.AdjustCodePage(m_CodePage);
		return result;
	}
// The RES is assumed to be a custom text resource
// index - contains the correct index
	if(!index || res.GetLength() < 3)
		return result;
	CString TypeName = res.Left(index);
	TypeName.MakeUpper();
	bool IsHTML = (TypeName == _T("HTML"));
	CString ResName = res.Right(res.GetLength() - index - 1);
//*******************************
	BYTE * pBuff = NULL;
	int nSize = 0;
	LPCWSTR ResType = TypeName;
	if(IsHTML)
		ResType = RT_HTML;
	if(GetResource(hInst, ResName, ResType, pBuff, nSize))
	{
		if(nSize < 1)
			return result;

		// Controlling the size of the loaded:
		if(m_nSizeLimit > 0 && m_nSizeLimit < nSize)
			nSize = m_nSizeLimit;
		// ***********************************

		pBuff = new BYTE[nSize];
		if(GetResource(hInst, ResName, ResType, pBuff, nSize))
			result.TakeFromANSI((LPCSTR)pBuff, nSize, m_CodePage);
		delete [] pBuff;
	}
//****************************
	return result;
}

CString CTOTxt::GetTextFromTag(CTag * tag)
{
	bool HasRes, HasFile;
	CString res = tag->GetAttrTextValue(attrRES, &HasRes);
	CString file = tag->GetAttrTextValue(attrFILE, &HasFile);

// *** ID support ****************************
	bool bHasID;
	long ID = tag->GetAttrIntValue(_T("ID"), &bHasID);
	if(bHasID)
	{
		if(!file.IsEmpty())
			file = m_pGlobalObject->GetValidFilePath(file);
		CString tmp = m_pGlobalObject->OnGetText(file, res, m_nSizeLimit, ID);
		tmp.AdjustCodePage(m_CodePage);
		return tmp;
	}
// *******************************************

	CString result;
	if(!HasRes && !HasFile)
		return result;

	if(HasFile && !HasRes)
		return GetTextFromFile(tag);

	if(!HasFile && HasRes)
		return GetTextFromRes(m_pGlobalObject->GetDefaultInstance(), res);

// Otherwise, we have both file and the resource specified

	file = m_pGlobalObject->GetValidFilePath(file);
	if(file.IsEmpty())
		return result;

#ifdef _MSLU
	HINSTANCE hInst;
	if(g_bNTCompitable)
		hInst = ::LoadLibraryEx(file, NULL, LOAD_LIBRARY_AS_DATAFILE);
	else
		hInst = ::LoadLibrary(file);
#else
	HINSTANCE hInst = ::LoadLibraryEx(file, NULL, LOAD_LIBRARY_AS_DATAFILE);
#endif
	
	if(!hInst)
		return result;
	result = GetTextFromRes(hInst, res);
	FreeLibrary(hInst);
	return result;
}

// *************************************************
// CTODocument implementation **********************
// *************************************************

CTODocument::CTODocument(CTag * tag, CGlobalObject * globalObject):CTagObject(tag)
{
// Gradient support: *********************************
	m_grad.Init(tag);
// ***************************************************

	m_nBorder = max(tag->GetAttrIntValue(attrBORDER), 0);

	bool bColorValid;
	m_BorderColor = tag->GetAttrColorValue(attrBC, bColorValid);
	if(!bColorValid)
		m_BorderColor = tag->GetAttrColorValue(attrBORDERCOLOR, bColorValid);

	if(!bColorValid)
		m_BorderColor = RGB(255, 0, 0);
	m_bgColor = tag->GetAttrColorValue(attrBGCOLOR, m_bBGColorValid);
	m_pGlobalObject = globalObject;
	m_strFile = tag->GetAttrTextValue(attrFILE);
	if(!m_strFile.IsEmpty())
		m_strFile = globalObject->GetValidFilePath(m_strFile);
	m_strParam = tag->GetAttrTextValue(attrPARAM);
	m_docSize.cx = max(tag->GetAttrIntValue(attrWIDTH), 0); // Using MAX, because negative values are not valid
	m_docSize.cy = max(tag->GetAttrIntValue(attrHEIGHT), 0);

	m_pBackground = CPicture::GetObjectFromTag(tag, NULL, globalObject);
}

CTODocument::~CTODocument()
{
	if(m_pBackground)
		delete m_pBackground;
}

CSize CTODocument::GetSize(HDC dc)
{
	m_docSize = m_pGlobalObject->GetDocumentSize(m_strFile, m_strParam, m_docSize.cx, m_docSize.cy);
	m_docSize.cx += m_nBorder * 2;
	m_docSize.cy += m_nBorder * 2;
	return m_docSize;
}

void CTODocument::OnDrawTag(HDC dc, int x, int y, int width, int height, int ZeroX, int ZeroY)
{
	if(m_docSize.IsEmpty()) // Do not draw a document with width or height set to zero.
		return;

	if(m_grad.m_bGFCValid)
		FillGradient(dc, m_bgColor, m_grad.m_GFC, x, y, x + width, y + height, m_grad.m_bGFCHorizontal);
	else
		if(m_bBGColorValid)
		{
			HBRUSH brush = CreateSolidBrush(m_bgColor);
			HGDIOBJ OldBrush = SelectObject(dc, brush);
			CRect rect(x, y, x + width, y + height);
			FillRect(dc, rect, brush);
			SelectObject(dc, OldBrush);
			DeleteObject(brush);
		}

	if(m_nBorder)
	{
		HBRUSH brush = CreateSolidBrush(m_BorderColor);
		HGDIOBJ OldBrush = SelectObject(dc, brush);
		RECT r1 = {x, y, x + m_nBorder, y + height};
		RECT r2 = {x, y, x + width, y + m_nBorder};
		RECT r3 = {x, y + height - m_nBorder, x + width, y + height};
		RECT r4 = {x + width - m_nBorder, y, x + width, y + height};
		FillRect(dc, &r1, brush);
		FillRect(dc, &r2, brush);
		FillRect(dc, &r3, brush);
		FillRect(dc, &r4, brush);
		SelectObject(dc, OldBrush);
		DeleteObject(brush);
	}
	if(m_pBackground)
		m_pBackground->Populate(dc, x + m_nBorder, y + m_nBorder, width - m_nBorder * 2, height - m_nBorder * 2);
	m_pGlobalObject->DrawDocument(m_strFile, m_strParam, dc, x + m_nBorder, y + m_nBorder, width - m_nBorder * 2, height - m_nBorder * 2);
}

// *************************************************
// CTOControl implementation ***********************
// *************************************************

CTOControl::CTOControl(CTag * tag, CGlobalObject * globalObject):CTagObject(tag)
{
	m_ID = tag->GetAttrIntValue(_T("ID"));
	m_Tag = *tag;
	SetTagAttr(TA_CONTROL);
	m_ControlSize.cx = max(tag->GetAttrIntValue(attrWIDTH), 0); // Using MAX, because negative values are not valid
	m_ControlSize.cy = max(tag->GetAttrIntValue(attrHEIGHT), 0);
	m_strName = tag->GetAttrTextValue(_T("NAME"));
	m_strParam = tag->GetAttrTextValue(attrPARAM);

	m_pGlobalObject = globalObject;
}

CSize CTOControl::GetSize(HDC dc)
{
	if(m_ID)
		return m_pGlobalObject->GetControlSize(m_ID, m_strName, m_ControlSize.cx, m_ControlSize.cy, m_strParam);
	return CSize(0, 0);
}

