/*
 * VXPLib COMPONENT
 *
 * Module: TagObject.cpp - Base class for all tag classes
 * (CTOxxx classes)
 *
 * Synopsis: This file contains implementation of class CTagObject
 * with properties and methods general for each tag class in the library.
 * It also implements the very core of the component's QHTML engine.
 *
 * Copyright (c) 2002, Vitaly Tomilov
 * Copyright (c) 2003-2005, Tooltips.Net L.L.C.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * Library support web site - http://www.tooltips.net
 */

#include "stdafx.h"
#include "TagObject.h"
#include "TableMatrix.h"

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

CTagObject::CTagObject(CTag * tag)
{
	m_pTable = NULL;

	m_TagAttr = TA_TEXT;		// assuming a simple text object
	m_bVisual = false;			// No default visualization
	m_NextLineJump = 0;			// No jumping to the next line
	m_bUnused = false;
	m_client.right = -1;		// To indicate that the object got no size
	m_FullRect.right = -1;

	m_uHAlignment = ALIGN_NONE;
	m_uVAlignment = ALIGN_NONE;

	m_MaxLine = 0;
}

// The important thing is that when calculating the size
// of objects alignment is not used. The alignment is important
// only at the stage of drawing, because it doesn't effect
// objects' sizes in any way.
// The calculation is to consist of two types of size measurements:
// For in-table elements, and for all the rest.
// When a table is found, it should calculate its result on its own.
CSize CTagObject::CalculateSize(HDC dc)
{
	m_Zero_X = 0;
	m_Zero_Y = 0;
	CSize size = GetSequenceSize(dc, NULL, &m_Zero_X, &m_Zero_Y);
	SetHRWidth(size.cx);
	return size;
}

CSize CTagObject::GetSize(HDC dc)
{
	// It is Ok for the abstract object to return like
	// this, because it is not drawn by default
	return CSize(0, 0);
}

void CTagObject::Draw(HDC dc, int x, int y)
{
// Whatever drawing is performed, all obects with m_bUnused set
// to true must be skipped.
// All clients are to be drawn within their m_client rectangle
	CTagObject * object = this;
	while(object)
	{
		if(!object->m_bUnused)
			object->OnDrawTag(dc, x + object->m_client.left, y + object->m_client.top, object->m_client.right, object->m_client.bottom, x, y);
		object = (CTagObject*)object->Next;
	}
	// Now we need to make sure that all column borders are drawn as well
	object = this;

	while(object)
	{
		if(!object->m_bUnused && object->m_TagAttr == TA_TD)
			object->OnDrawTDBorders(dc, x, y);
		object = (CTagObject*)object->Next;
	}
}

void CTagObject::SetTagAttr(int attr)
{
	m_TagAttr = attr;
}

// Sets how many lines are to be skipped for the current object
// Normally, it is 1 for BR, 2 for HR, and 0 for all the rest
// However, such objects as BR and HR can require even more
// lines to be skipped.
void CTagObject::SetNextLineJump(int nLines)
{
	m_NextLineJump = nLines;
}

// This function searches for the end of the table beginning
// from the next object. When another table is met, it is
// ignored. A counter used to throw away inner tables.
// If not found, NULL is returned.
CTagObject * CTagObject::GetTableEnd()
{
	CTagObject * A = (CTagObject*)Next;
	int end_counter = 0;
	while(A)
	{
		if(A->m_TagAttr == TA_TABLE_END)
		{
			if(end_counter)
				end_counter --;
			else
				return A;
		}
		else
			if(A->m_TagAttr == TA_TABLE)
				end_counter ++;
		A = (CTagObject*)A->Next;
	}
	return NULL;
}

// This function must return size of the table, where
// this = beginning of the table; end = the end of the table
CSize CTagObject::GetTableSize(HDC dc, CTagObject *start, CTagObject *end, int * zero_X, int * zero_Y)
{
	CSize result; // The resulting size

// Do not analyze tables, which have no elements in them
	if(start->Next == end)
		return result;

// *****************************************************
// Step1: Create a matrix of cells

	// First of all, we do not create anything big
	// right in the procedure stack.
	CTableMatrix & tm = *new CTableMatrix(start, end);
	CSize matr_size = tm.GetMatrixSize();
	if(matr_size.IsEmpty()) // If there are no rows/columns in the table
	{
		delete &tm;
		return result;
	}

// Step2: Use GetSequenceSize() with the second parameter - 
// last object limiter

// The height of each <TR> and the width of each <TD> is
// to be taken here into account later
	int height = 0;
	int last_zero_X = *zero_X;
	int last_zero_Y = *zero_Y;
	int cellpadding = start->GetCellPadding();
	int cellspacing = start->GetCellSpacing();
	int i, k;

	for(i = 0;i < matr_size.cy;i ++)
	{
		int width = 0;
		int max_height = 0;
		CSize tmp;
		for(k = 0;k < matr_size.cx;k ++)
		{
			CMatrixCell * cell = tm.GetCell(k, i);
			*zero_X += tmp.cx;
			if(!cell->IsEmpty())
				tmp = cell->m_pFirst->GetSequenceSize(dc, (CTagObject*)cell->m_pLast->Next, zero_X, zero_Y);
			else
				tmp.Zero();
			tmp.cx += cellpadding * 2;
			tmp.cy += cellpadding * 2;
			cell->m_RealContents = tmp;
			tmp.AdjustToMax(cell->m_MinSize); // Take into account the minimum cell size
			//*** Drawing support *****************************************
			cell->m_contentSize = tmp; // Store contents size for each cell
			//*************************************************************
			width += tmp.cx;
			if(tmp.cy > max_height)
				max_height = tmp.cy;
		}
		// *** Drawing support **************************************
		// Storing real height for each column in the current row
		for(k = 0;k < matr_size.cx;k ++)
			tm.GetCell(k, i)->m_cellSize.cy = max_height;
		// **********************************************************
		height += max_height;
		*zero_Y += max_height;
		*zero_X = last_zero_X;
	}

	*zero_Y = last_zero_Y;

// The last step, we need to align for each cell its zero_X
// in order to start from the width of the widest column

// Idea: for each object, when calling SetClientRect() we also mark
// as Unaligned. Above, into each cell we write the cell's size
// Then from start till the end we'll align all objects.

// Each cell may also contain its real size!!!

// We must set the real width into each cell
// In order to do this we need to know the width of each column
	for(i = 0;i < matr_size.cx;i ++) // For each column
	{
		int max_width = 0;
		for(k = 0;k < matr_size.cy;k ++) // For each row
		{
			CMatrixCell * cell = tm.GetCell(i, k);
			int width = cell->m_contentSize.cx;
			if(width > max_width)
				max_width = width;
		}
		// Now we know width for column i
		for(k = 0;k < matr_size.cy;k ++) // For each row
			tm.GetCell(i, k)->m_cellSize.cx = max_width;
	}

// *******************************************************************
// Now in each cell we have its real size and the size of its contents,
// which is anough for objects alignment
// And we need to recalculate sizes of all objects:

	int border = start->GetBorderSize();

// Step1: Change X and Y for all objects:

	ShiftObjects(&tm, border, cellpadding, cellspacing);

// Step2: Recalculate size by X

	int total_width = 0;
	for(i = 0;i < matr_size.cx;i ++) // For all columns
		total_width += tm.GetCell(i, 0)->m_cellSize.cx;

	total_width += cellspacing * (matr_size.cx - 1) + border * 2;
	height += cellspacing * (matr_size.cy - 1) + border * 2;

	tm.SetBorderPos(last_zero_X, last_zero_Y, total_width, height, border, cellspacing);

	result.SetSize(total_width, height);

	start->m_client.right = total_width;	// Important for alignment
	start->m_client.bottom = height;		// Important for alignment

	AlignObjects(&tm, cellpadding);

// ********************************************

	int ShiftX = 0;
	int ShiftY = 0;
	int MaxRow = tm.GetMaxColsRow();

	for(i = 0;i < matr_size.cy;i ++)
	{
		for(k = 0;k < matr_size.cx;k ++)
		{
			CMatrixCell * cell = tm.GetCell(k, i);
			if(cell->m_pTD)
			{
				cell->m_pTD->SetFullRect(last_zero_X + border + ShiftX + cellspacing * k, border + last_zero_Y + ShiftY, cell->m_cellSize.cx, cell->m_cellSize.cy);
				cell->m_pTR->SetFullRect(last_zero_X + border, border + last_zero_Y + ShiftY, total_width - border * 2, cell->m_cellSize.cy);
				ShiftX += tm.GetCell(k, MaxRow)->m_cellSize.cx;
			}
		}
		ShiftX = 0;
		ShiftY += tm.GetCell(0, i)->m_cellSize.cy + cellspacing;
	}

// *********************************************

	delete &tm;
	return result;
}

CSize CTagObject::GetSequenceSize(HDC dc, CTagObject * pLastObject, int * zero_X, int * zero_Y)
{
	CTagObject * object = this;

	int MaxHeight = 0;
	int MaxWidth = 0;

	int Current_X = 0;
	int Current_Y = 0;

	CTagObject * shift = this;

	//************
	int MaxTopHeight = 0; // Maximum height of the text elemens, excluding bottom symbol endings
	int CurrentTopHeight = 0;
	//************
	while(object != pLastObject)
	{
		CSize size;
		CTagObject * nextObject;
		if(object->m_TagAttr == TA_TABLE) // If we found a table
		{
			nextObject = object->GetTableEnd();
			if(!nextObject) // If no table end found, we ignore all the rest, and mark all the rest as unused
			{
				while(object)
				{
					object->SetUnused();
					object = (CTagObject*)object->Next;
				}
				break;
			}

			int last_zero_X = *zero_X;
			int last_zero_Y = *zero_Y;

			*zero_X += Current_X;
			*zero_Y += Current_Y;
			size = GetTableSize(dc, object, nextObject, zero_X, zero_Y);
			size.AdjustToMax(object->GetForcedSize());
			*zero_X = last_zero_X;
			*zero_Y = last_zero_Y;

			nextObject = (CTagObject*)nextObject->Next;
		}
		else
		{
			nextObject = (CTagObject*)object->Next;
			size = object->GetSize(dc); // The key call
			object->SetClientRect(*zero_X + Current_X, *zero_Y + Current_Y, size.cx, size.cy);
			CurrentTopHeight = object->GetTopHeight(dc);
		}
		if(object->m_NextLineJump)
		{
			if(object->GetTagAttr() == TA_BR && Current_X)
				size.cy = 0;

			object->m_client.left = *zero_X;
			object->m_client.top = *zero_Y + Current_Y + MaxHeight;
			object->m_client.bottom = size.cy;

			Current_X = 0;
			Current_Y += MaxHeight + size.cy;

			//*** Setting m_MaxLine for all elements in the line ***
			CTagObject * A = shift;
			int skip = 0;
			while(1)
			{
				if(A->GetTagAttr() == TA_TABLE)
					skip ++;
				if(A->GetTagAttr() == TA_TABLE_END)
					skip --;
				if(!skip)
					A->m_MaxLine = MaxTopHeight;
				if(A == object)
					break;
				A = (CTagObject*)A->Next;
			}
			shift = object;

			MaxTopHeight = 0;
			//*** End of setting m_MaxLine for all elements in the line ***

			MaxHeight = 0;
		}
		else
		{
			Current_X += size.cx;
			MaxWidth = max(MaxWidth, Current_X);
			MaxHeight = max(MaxHeight, size.cy);

			//*** Building up MaxTopHeight, if necessary ***
			MaxTopHeight = max(CurrentTopHeight, MaxTopHeight);
		}
		
		//*** Setting m_MaxLine for all elements in the line ***
		if(nextObject == pLastObject)
		{
			CTagObject * A = shift;
			int skip = 0;
			while(1)
			{
				if(A->GetTagAttr() == TA_TABLE)
					skip ++;
				if(A->GetTagAttr() == TA_TABLE_END)
					skip --;
				if(!skip)
					A->m_MaxLine = MaxTopHeight;
				if(A == object)
					break;
				A = (CTagObject*)A->Next;
			}
			shift = object;
		}
		//*** End of setting m_MaxLine for all elements in the line ***

		object = nextObject;
	}
	return CSize(MaxWidth, Current_Y + MaxHeight);
}

int CTagObject::GetTagAttr()
{
	return m_TagAttr;
}

void CTagObject::SetUnused()
{
	m_bUnused = true;
}

void CTagObject::SetClientRect(int x, int y, int width, int height)
{
	m_client.left = x;
	m_client.top = y;
	m_client.right = width;
	m_client.bottom = height;
}

// This function is to shift the X and Y coordinates for all
// objects in the matrix
void CTagObject::ShiftObjects(CTableMatrix *matr, int border, int cellpadding, int cellspacing)
{
	CSize size = matr->GetMatrixSize();
	int shiftY = cellpadding + border;
	for(int i = 0;i < size.cy;i ++)
	{
		int shiftX = cellpadding + border;
		for(int k = 0;k < size.cx;k ++)
		{
			CMatrixCell * cell = matr->GetCell(k, i);
			if(!cell->IsEmpty())
			{
				CTagObject * A = cell->m_pFirst;
				while(A != cell->m_pLast)
				{
					A->m_client.left += shiftX;
					A->m_client.top += shiftY;
					//*******************************
					if(A->m_FullRect.right != -1)
					{
						A->m_FullRect.left += shiftX;
						A->m_FullRect.top += shiftY;
					}
					//*******************************
					A = (CTagObject*)A->Next;
				}
				A->m_client.left += shiftX;
				A->m_client.top += shiftY;
				//*******************************
				if(A->m_FullRect.right != -1)
				{
					A->m_FullRect.left += shiftX;
					A->m_FullRect.top += shiftY;
				}
				//*******************************
			}
			shiftX += cell->m_cellSize.cx - cell->m_contentSize.cx + cellspacing;
		}
		shiftY += cellspacing;
	}
}

// This function must align all objects in cells
// Using the cell's parameters m_pTD and m_pTR
void CTagObject::AlignObjects(CTableMatrix *matr, int cellpadding)
{
	CSize size = matr->GetMatrixSize();
	for(int i = 0;i < size.cy;i ++)
	{
		for(int k = 0;k < size.cx;k ++)
		{
			CMatrixCell * cell = matr->GetCell(k, i);
			if(!cell->IsEmpty())
			{
				CSize cont_size = cell->m_RealContents;
				CTagObject * A = cell->m_pFirst;
				do
				{
					// Step1: Calculating the width of the first portion of objects
					// before the end/or before <BR>
					CTagObject * tmp = A;
					int sx = 0;
					while(!A->m_NextLineJump)
					{
						sx += A->m_client.right;
						if(A->GetTagAttr() == TA_TABLE)
							A = A->GetTableEnd();
						if(A == cell->m_pLast)
							break;
						A = (CTagObject*)A->Next;
					}
					CTagObject * B = tmp;
					// Horizontal alignment:

					int alignment = cell->m_pTD->GetHAlignment();
					if(alignment == ALIGN_NONE)
						alignment = cell->m_pTR->GetHAlignment();
					int skip=0;
					do
					{
						// <HR> tags are not aligned horizontally,
						// but only vertically
						int t = B->GetTagAttr();
						if(t == TA_TABLE)
							skip ++;
						else
							if(t == TA_TABLE_END)
								skip --;

						if(B->GetTagAttr() != TA_HR || skip)
						{
							switch(alignment)
							{
							case ALIGN_CENTER:
								{
									int shiftX = (cell->m_cellSize.cx - sx)/2 - cellpadding;
									B->m_client.left += shiftX;
									if(B->m_FullRect.right != -1)
										B->m_FullRect.left += shiftX;
									break;
								}
							case ALIGN_RIGHT:
								{
									int shiftX = (cell->m_cellSize.cx - sx) - cellpadding * 2;
									B->m_client.left += shiftX;
									if(B->m_FullRect.right != -1)
										B->m_FullRect.left += shiftX;
									break;
								}
							default:
								break;
							}
						}
						else
							if(!skip)
								B->m_client.right = cell->m_cellSize.cx - cellpadding * 2;

						if(B == cell->m_pLast || B == A)
							break;
						else
							B = (CTagObject*)B->Next;
					}
					while(1);

					// Vertical alignment:

					B = tmp;
					alignment = cell->m_pTD->GetVAlignment();
					if(alignment == ALIGN_NONE)
						alignment = cell->m_pTR->GetVAlignment();
					do
					{
						switch(alignment)
						{
						case ALIGN_MIDDLE:
							{
								int shiftY = (cell->m_cellSize.cy - cont_size.cy)/2;
								B->m_client.top += shiftY;
								if(B->m_FullRect.right != -1)
									B->m_FullRect.top += shiftY;
								break;
							}
						case ALIGN_BOTTOM:
							{
								int shiftY = (cell->m_cellSize.cy - cont_size.cy);
								B->m_client.top += shiftY;
								if(B->m_FullRect.right != -1)
									B->m_FullRect.top += shiftY;
								break;
							}
						default:
							break;
						}
						if(B == cell->m_pLast || B == A)
							break;
						else
							B = (CTagObject*)B->Next;
					}
					while(1);

					if(A == cell->m_pLast)
						break;

					A = (CTagObject*)A->Next;
				}
				while(1);
			}

		}
	}
}

int CTagObject::GetCellPadding()
{
	return 0;
}

int CTagObject::GetCellSpacing()
{
	return 0;
}

void CTagObject::SetFullRect(int x, int y, int width, int height)
{
	m_FullRect.left = x;
	m_FullRect.top = y;
	m_FullRect.right = width;
	m_FullRect.bottom = height;
}

int CTagObject::GetHAlignment()
{
	return m_uHAlignment;
}

int CTagObject::GetVAlignment()
{
	return m_uVAlignment;
}

void CTagObject::SetHAlignment(int HAlignment)
{
	m_uHAlignment = HAlignment;
}

void CTagObject::SetVAlignment(int VAlignment)
{
	m_uVAlignment = VAlignment;
}

int CTagObject::TextIntoAlignment(LPCTSTR text)
{
	if(!text)
		return ALIGN_NONE;
	CString value(text);
	value.MakeUpper();
	if(value == _T("CENTER"))
		return ALIGN_CENTER;
	if(value == _T("TOP"))
		return ALIGN_TOP;
	if(value == _T("BOTTOM"))
		return ALIGN_BOTTOM;
	if(value == _T("LEFT"))
		return ALIGN_LEFT;
	if(value == _T("RIGHT"))
		return ALIGN_RIGHT;
	if(value == _T("MIDDLE"))
		return ALIGN_MIDDLE;

	return ALIGN_NONE;
}

void CTagObject::SetHRWidth(int width)
{
	CTagObject * object = this;
	int skip = 0;
	while(object)
	{
		if(!object->IsUnused())
		{
			switch(object->GetTagAttr())
			{
			case TA_TABLE:
				{
					skip ++;
					break;
				}
			case TA_TABLE_END:
				{
					skip --;
					if(skip < 0)
						skip = 0;
					break;
				}
			case TA_HR:
				{
					if(!skip)
						object->m_client.right = width;
					break;
				}
			default:
				break;

			}
		}
		object = (CTagObject*)object->Next;
	}
}
