/*
 * VXPLib COMPONENT
 *
 * Module: TagObject.h - Base class for all tag classes
 * (CTOxxx classes)
 *
 * Synopsis: This file contains declaration of class CTagObject
 * that implements properties and methods general for each tag
 * class in the library.
 * It also implements the very core of the component's QHTML engine.
 *
 * Copyright (c) 2002, Vitaly Tomilov
 * Copyright (c) 2003-2005, Tooltips.Net L.L.C.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * Library support web site - http://www.tooltips.net
 */

#if !defined(AFX_TAGOBJECT_H__9D564868_2AAF_4F13_BA92_F0234CA5E99C__INCLUDED_)
#define AFX_TAGOBJECT_H__9D564868_2AAF_4F13_BA92_F0234CA5E99C__INCLUDED_

#if _MSC_VER > 1000
#pragma once
#endif // _MSC_VER > 1000

#include "HCollection.h"
#include "Tag.h"
#include "Size.h"

// Alignment parameters are applicable to each object
#define ALIGN_NONE		0
#define ALIGN_LEFT		1
#define ALIGN_RIGHT		2
#define ALIGN_TOP		3
#define ALIGN_BOTTOM	4
#define ALIGN_CENTER	5
#define ALIGN_MIDDLE	6

class CTableMatrix;

// The key base class to hold all necessary information
// about one tag
class CTagObject:public HCollection
{
	friend class CTableMatrix;
public:
	// Constructor:
	CTagObject(CTag * tag);
	// Destructor:
	virtual ~CTagObject(){};

	int GetVAlignment();
	int GetHAlignment();
	void SetFullRect(int x, int y, int width, int height);

	inline bool IsUnused(){return m_bUnused;};

	void SetUnused();
	int GetTagAttr();

	void SetClientRect(int x, int y, int width, int height);

	// Derived objects must return clear text
	virtual CString GetText(){return CString();};

	// Any object that visually splits text must
	// return true, except text itself.
	// This method is to support Speach SDK.
	virtual bool IsTextSplitter(){return false;};

	// Called for the first object in the collection
	// Must return total size of the tooltip
	CSize CalculateSize(HDC dc);
	
	// Must return the object size
	virtual CSize GetSize(HDC dc);

	// Some objects, like <table> and <td> may force the object to
	// have at least the size specified by "width" and "height"
	// Such objects must override this virtual function and
	// return the values they were passed.
	virtual CSize GetForcedSize(){return CSize();};

	// Some objects, like <table> support the "border" attribute
	// to set width and height of plus to its size.
	virtual int GetBorderSize(){return 0;};

	// Special for <TD> method called to draw
	// <TD> column border lines
	virtual void OnDrawTDBorders(HDC dc, int ZeroX, int ZeroY){};

	// This method is to be overridden by the text tag to return
	// the size of the text without bottom endings
	virtual int GetTopHeight(HDC dc){return 0;}

	void Draw(HDC dc, int x, int y);

	CRect m_client;	// Contains x,y,width and height for all objects' drawing

protected:

	// This method, when overridden is supposed to draw something
	virtual void OnDrawTag(HDC dc, int x, int y, int width, int height, int ZeroX, int ZeroY){};

	void SetNextLineJump(int nLines);
	void SetTagAttr(int attr);
	int m_TagAttr;		// Set by descendant to one of TA_xxx values
	bool m_bVisual;		// Is true when the object can be drawn
	int m_NextLineJump;	// Set to how many times the tag forces to jump
						// on the next line with the step equal to the current
						// font size.
	
	CTagObject * m_pTable;	// Can be set for some objects to point on the
							// table tag to which this tag belongs

private:
	void SetHRWidth(int width);
	void AlignObjects(CTableMatrix *matr, int cellpadding);
	void ShiftObjects(CTableMatrix * matr, int border, int cellpadding, int cellspacing);
	CSize GetSequenceSize(HDC dc, CTagObject * pLastObject, int * zero_X, int * zero_Y);
	CSize GetTableSize(HDC dc, CTagObject * start, CTagObject * end, int * zero_X, int * zero_Y);
	CTagObject * GetTableEnd();
		
	bool m_bUnused;	// Set to true for objects, which were ignored while
					// the tooltip size was calculated, and must be ignored
					// during the drawing/initialization procedure.
					// It relates to tags with wrong context or else.

	int m_Zero_X;	// Current zero x
	int m_Zero_Y;	// Current zero y

	int m_uHAlignment;	// Horizontal alignment of the object
	int m_uVAlignment;	// Vertical alignment of the object

protected:
	int TextIntoAlignment(LPCTSTR text);
	void SetVAlignment(int VAlignment);
	void SetHAlignment(int HAlignment);
	virtual int GetCellSpacing();
	virtual int GetCellPadding();
	
	CRect m_FullRect; // Contains x,y,width and height for TD and TR tags

	int m_MaxLine;	// The height of the highest text element in the line
					// excluding its bottom endings
};

#define TAG_OBJECT(T)\
class T:public CTagObject\
{\
public:\
	T(CTag * tag);\
};

#endif // !defined(AFX_TAGOBJECT_H__9D564868_2AAF_4F13_BA92_F0234CA5E99C__INCLUDED_)
