/*
 * VXPLib COMPONENT
 *
 * Module: VXPFont.cpp - Implementation of class CVXPFont
 *
 * Synopsis: Class CVXPFont implements methods and properties
 * of interface IVXPFont.
 *
 * Copyright (c) 2002, Vitaly Tomilov
 * Copyright (c) 2003-2005, Tooltips.Net L.L.C.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * Library support web site - http://www.tooltips.net
 */

#include "stdafx.h"
#include "VXPLib.h"
#include "VXPFont.h"
#include "ocidl.h"	// Added by ClassView

/////////////////////////////////////////////////////////////////////////////
// CVXPFont

CVXPFont::CVXPFont()
{
	SetDefaultParameters();
	m_hFont = NULL;
}

CVXPFont::~CVXPFont()
{
	if(m_hFont)
		DeleteObject(m_hFont);
}

void CVXPFont::SetDefaultParameters()
{
	SetSystemFont(fontMessage);

// Setting independent parameters
	m_Color = RGB(0, 0, 0);			// Black color
	m_bkColor = RGB(255, 255, 255);	// White color
	m_bTransparent = true;
}

STDMETHODIMP CVXPFont::SetFromSystemFont(VXPSysFont FontIndex, VARIANT_BOOL * success)
{
	VXPLIB_BEGIN

	*success = SetSystemFont(FontIndex);
	
	VXPLIB_END
}

BOOL CVXPFont::SetSystemFont(VXPSysFont FontIndex)
{
// Reading system font parameters:
	NONCLIENTMETRICS ncm;
	memset(&ncm, 0, sizeof(ncm));
	ncm.cbSize = sizeof(ncm);
	::SystemParametersInfo(SPI_GETNONCLIENTMETRICS, 0, (PVOID) &ncm, 0);

	switch(FontIndex)
	{
	case fontCaption:
		{
			m_LogFont = ncm.lfCaptionFont;
			m_bChanged = true;
			break;
		}
	case fontSmallCaption:
		{
			m_LogFont = ncm.lfSmCaptionFont;
			m_bChanged = true;
			break;
		}
	case fontMenu:
		{
			m_LogFont = ncm.lfMenuFont;
			m_bChanged = true;
			break;
		}
	case fontStatusBar:
		{
			m_LogFont = ncm.lfStatusFont;
			m_bChanged = true;
			break;
		}
	case fontMessage:
		{
			m_LogFont = ncm.lfMessageFont;
			m_bChanged = true;
			break;
		}
	default:
		{
			return FALSE;
		}
	}

	return TRUE;
}

void CVXPFont::UpdateFont()
{
	if(!m_bChanged)	
		return;

	m_bChanged = false;

	if(m_hFont)
	{
		DeleteObject(m_hFont);
		m_hFont = NULL;	// Just in case we fail to create the new font
	}

	m_hFont = CreateFontIndirect(&m_LogFont);
}

STDMETHODIMP CVXPFont::SetDefaults()
{
	VXPLIB_BEGIN

	SetDefaultParameters();

	VXPLIB_END
}

STDMETHODIMP CVXPFont::InterfaceSupportsErrorInfo(REFIID riid)
{
	static const IID* arr[] = 
	{
		&IID_IVXPFont
	};
	for (int i=0; i < sizeof(arr) / sizeof(arr[0]); i++)
	{
		if (InlineIsEqualGUID(*arr[i],riid))
			return S_OK;
	}
	return S_FALSE;
}

STDMETHODIMP CVXPFont::get_Transparent(VARIANT_BOOL *pVal)
{
	VXPLIB_BEGIN

	*pVal = m_bTransparent;
	return S_OK;

	VXPLIB_END
}

STDMETHODIMP CVXPFont::put_Transparent(VARIANT_BOOL newVal)
{
	VXPLIB_BEGIN

	m_bTransparent = (newVal != 0);
	
	VXPLIB_END
}

STDMETHODIMP CVXPFont::get_Name(BSTR * pVal)
{
	VXPLIB_BEGIN

	CString name(m_LogFont.lfFaceName);
	*pVal = name.SysAllocString();
	
	VXPLIB_END
}

STDMETHODIMP CVXPFont::put_Name(BSTR newVal)
{
	VXPLIB_BEGIN

	CString name(newVal);
	name.TrimLeft();
	name.TrimRight();
	_copynstr(m_LogFont.lfFaceName, name, min(name.GetLength(), LF_FACESIZE));

	m_bChanged = true;

	VXPLIB_END
}

STDMETHODIMP CVXPFont::get_hFont(long *pVal)
{
	VXPLIB_BEGIN

	UpdateFont();
	*pVal = (long)m_hFont;
	
	VXPLIB_END
}

STDMETHODIMP CVXPFont::get_Size(short *pVal)
{
	VXPLIB_BEGIN

	HDC dc = GetDC(NULL);
	*pVal = -MulDiv(m_LogFont.lfHeight, 72, GetDeviceCaps(dc, LOGPIXELSY));
	ReleaseDC(NULL, dc);

	VXPLIB_END
}

STDMETHODIMP CVXPFont::put_Size(short newVal)
{
	VXPLIB_BEGIN

	HDC dc = GetDC(NULL);
	m_LogFont.lfHeight = -MulDiv(newVal, GetDeviceCaps(dc, LOGPIXELSY), 72);
	ReleaseDC(NULL, dc);
	m_LogFont.lfWidth = 0;

	VXPLIB_END
}

STDMETHODIMP CVXPFont::get_Weight(VXPFontWeight *pVal)
{
	VXPLIB_BEGIN

	*pVal = (VXPFontWeight)m_LogFont.lfWeight;
	
	VXPLIB_END
}

STDMETHODIMP CVXPFont::put_Weight(VXPFontWeight newVal)
{
	VXPLIB_BEGIN

	m_LogFont.lfWeight = newVal;
	
	VXPLIB_END
}

STDMETHODIMP CVXPFont::get_Italic(VARIANT_BOOL *pVal)
{
	VXPLIB_BEGIN

	*pVal = m_LogFont.lfItalic;
	
	VXPLIB_END
}

STDMETHODIMP CVXPFont::put_Italic(VARIANT_BOOL newVal)
{
	VXPLIB_BEGIN

	m_LogFont.lfItalic = newVal?TRUE:FALSE;
	
	VXPLIB_END
}

STDMETHODIMP CVXPFont::get_Underlined(VARIANT_BOOL *pVal)
{
	VXPLIB_BEGIN

	*pVal = m_LogFont.lfUnderline;
	
	VXPLIB_END
}

STDMETHODIMP CVXPFont::put_Underlined(VARIANT_BOOL newVal)
{
	VXPLIB_BEGIN

	m_LogFont.lfUnderline = newVal?TRUE:FALSE;
	
	VXPLIB_END
}

STDMETHODIMP CVXPFont::get_StrikeOut(VARIANT_BOOL *pVal)
{
	VXPLIB_BEGIN

	*pVal = m_LogFont.lfStrikeOut;
	
	VXPLIB_END
}

STDMETHODIMP CVXPFont::put_StrikeOut(VARIANT_BOOL newVal)
{
	VXPLIB_BEGIN

	m_LogFont.lfStrikeOut = newVal?TRUE:FALSE;
	
	VXPLIB_END
}

STDMETHODIMP CVXPFont::get_Orientation(short *pVal)
{
	VXPLIB_BEGIN

	*pVal = (short)m_LogFont.lfOrientation;
	
	VXPLIB_END
}

STDMETHODIMP CVXPFont::put_Orientation(short newVal)
{
	VXPLIB_BEGIN

	m_LogFont.lfOrientation = newVal;
	
	VXPLIB_END
}

STDMETHODIMP CVXPFont::get_Escapement(short *pVal)
{
	VXPLIB_BEGIN

	*pVal = (short)m_LogFont.lfEscapement;
	
	VXPLIB_END
}

STDMETHODIMP CVXPFont::put_Escapement(short newVal)
{
	VXPLIB_BEGIN

	m_LogFont.lfEscapement = newVal;
	
	VXPLIB_END
}

STDMETHODIMP CVXPFont::get_CharSet(VXPCharSet * pVal)
{
	VXPLIB_BEGIN

	*pVal = (VXPCharSet)m_LogFont.lfCharSet;
	
	VXPLIB_END
}

STDMETHODIMP CVXPFont::put_CharSet(VXPCharSet newVal)
{
	VXPLIB_BEGIN

	m_LogFont.lfCharSet = (BYTE)newVal;
	
	VXPLIB_END
}

STDMETHODIMP CVXPFont::get_OutPrecision(short *pVal)
{
	VXPLIB_BEGIN

	*pVal = m_LogFont.lfOutPrecision;
	
	VXPLIB_END
}

STDMETHODIMP CVXPFont::put_OutPrecision(short newVal)
{
	VXPLIB_BEGIN

	m_LogFont.lfOutPrecision = (BYTE)newVal;
	
	VXPLIB_END
}

STDMETHODIMP CVXPFont::get_ClipPrecision(short *pVal)
{
	VXPLIB_BEGIN

	*pVal = m_LogFont.lfClipPrecision;
	
	VXPLIB_END
}

STDMETHODIMP CVXPFont::put_ClipPrecision(short newVal)
{
	VXPLIB_BEGIN

	m_LogFont.lfClipPrecision = (BYTE)newVal;
	
	VXPLIB_END
}


STDMETHODIMP CVXPFont::get_Quality(short *pVal)
{
	VXPLIB_BEGIN

	*pVal = m_LogFont.lfQuality;
	
	VXPLIB_END
}

STDMETHODIMP CVXPFont::put_Quality(short newVal)
{
	VXPLIB_BEGIN

	m_LogFont.lfQuality = (BYTE)newVal;
	
	VXPLIB_END
}

STDMETHODIMP CVXPFont::get_PitchAndFamily(short *pVal)
{
	VXPLIB_BEGIN

	*pVal = m_LogFont.lfPitchAndFamily;
	
	VXPLIB_END
}

STDMETHODIMP CVXPFont::put_PitchAndFamily(short newVal)
{
	VXPLIB_BEGIN

	m_LogFont.lfPitchAndFamily = (BYTE)newVal;
	
	VXPLIB_END
}

STDMETHODIMP CVXPFont::get_Color(OLE_COLOR *pVal)
{
	VXPLIB_BEGIN

	*pVal = m_Color;
	
	VXPLIB_END
}

STDMETHODIMP CVXPFont::put_Color(OLE_COLOR newVal)
{
	VXPLIB_BEGIN

	m_Color = newVal;
	
	VXPLIB_END
}

STDMETHODIMP CVXPFont::get_BackColor(OLE_COLOR *pVal)
{
	VXPLIB_BEGIN

	*pVal = m_bkColor;
	
	VXPLIB_END
}

STDMETHODIMP CVXPFont::put_BackColor(OLE_COLOR newVal)
{
	VXPLIB_BEGIN

	m_bkColor = newVal;
	
	VXPLIB_END
}

STDMETHODIMP CVXPFont::SetFromHandle(long hFont, VARIANT_BOOL * success)
{
	VXPLIB_BEGIN

	*success = GetObject((HGDIOBJ)hFont, sizeof(LOGFONT), &m_LogFont)?TRUE:FALSE;
	
	m_bChanged = true;
	
	VXPLIB_END
}
