/*
 * VXPLib COMPONENT
 *
 * Module: VXPQuickHTML.cpp - Implementation of class CVXPQuickHTML
 *
 * Synopsis: Class CVXPQuickHTML implements methods and properties
 * of interface IVXPQuickHTML.
 *
 * Copyright (c) 2002, Vitaly Tomilov
 * Copyright (c) 2003-2005, Tooltips.Net L.L.C.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * Library support web site - http://www.tooltips.net
 */

#include "stdafx.h"
#include "VXPLib.h"
#include "VXPQuickHTML.h"
#include "TagLib.h"
#include "TOFont.h"
#include "AboutBox.h"

/////////////////////////////////////////////////////////////////////////////
// CVXPQuickHTML

CVXPQuickHTML::CVXPQuickHTML()
{
	m_bRecalculating = false;

	m_pLastTag = NULL;
	m_Exception = 0;

	m_pTags = NULL;
	m_bTransparent = false;
	m_MemoryBMP = NULL;
	m_bSourceChanged = false;
	
	m_bSetupBody = false;
	m_bSetupSound = false;

	m_pBackground = NULL;

	m_nControls = 0;
	m_MinID = 0;
	m_MaxID = 0;
}

CVXPQuickHTML::~CVXPQuickHTML()
{
	if(m_MemoryBMP)
		DeleteObject(m_MemoryBMP);

	if(m_pBackground)
		delete m_pBackground;

	DELETE_COLLECTION(m_pTags)
}

HRESULT CVXPQuickHTML::FinalConstruct()
{
	::CoCreateInstance(CLSID_VXPSettings, NULL, CLSCTX_ALL, IID_IVXPSettings, (void**)&m_pSettings);

	m_pSettings->get_margins(&m_pMargins);
	m_pSettings->get_font(&m_pFont);

	IVXPSound * pSound = NULL;
	m_pSettings->get_bgsound(&pSound);
	if(pSound)
	{
		pSound->QueryInterface(CLSID_VXPSound, (void**)&m_pSoundClass);
		pSound->Release();
	}

	return S_OK;
}

HRESULT CVXPQuickHTML::FinalRelease()
{
	m_pSoundClass->Release();
	m_pMargins->Release();
	m_pFont->Release();

	m_pSettings->Release();

	return S_OK;
}

STDMETHODIMP CVXPQuickHTML::InterfaceSupportsErrorInfo(REFIID riid)
{
	static const IID* arr[] = 
	{
		&IID_IVXPQuickHTML
	};
	for (int i=0; i < sizeof(arr) / sizeof(arr[0]); i++)
	{
		if (InlineIsEqualGUID(*arr[i],riid))
			return S_OK;
	}
	return S_FALSE;
}

STDMETHODIMP CVXPQuickHTML::SetSourceText(BSTR source)
{
	VXPLIB_BEGIN

	m_text.TakeFromBSTR(source, m_BodyCodePage);

	m_bSourceChanged = true;

	m_text.Replace('\n', ' ');
	m_text.Replace('\r', ' ');
	m_text.Replace('\t', ' ');
	m_text.ShrinkEffective();

	VARIANT_BOOL param = m_text.IsEmpty()?TRUE:FALSE;
	
	Fire_OnSourceChanged(param);

	VXPLIB_END
}

// Even when the source hasn't changed, the contents still can be different
// due to dynamic nature of the contents, like <XML> things,
// which query for the contents dynamically.
// Therefore whenever we get into this procedure we always must process
// the whole contents over again.
// However, apart from this throughout the object we will be doing it
// only when the source change directly. For this reason, it is responsibility
// of the client to call this method when he suspects that the dynamic contents
// of the source has changed.
// This is why we also call Fire_OnRecalculate, suspecting that
// the client expects a change in the dynamic contents of the source.
STDMETHODIMP CVXPQuickHTML::RecalculateLayout(VARIANT_BOOL * success)
{
	VXPLIB_BEGIN

	*success = FALSE;

	if(!m_bRecalculating)
	{
		m_bRecalculating = true;

		Fire_OnRecalculate(TRUE, m_bSourceChanged);

		// Resetting the body and the sound tags: <BODY>, <BGSOUND>
		m_bSetupBody = false;
		m_bSetupSound = false;

		m_pSettings->SetDefaults();

		// Updating Windows System palette:
		CTag::UpdateSysColors();

		// In responce to this the derived interface or a client application
		// is supposed to retrieve IVXPSettings contained in this object
		// and populate it with default parameters.
		Fire_OnSetDefaults();

		ResetState();

		// Deleting the background picture object
		// before recalculating the size, because
		// SetupBody, which creates the background picture
		// will be called during the size recalculation:
		if(m_pBackground)
		{
			delete m_pBackground;
			m_pBackground = NULL;
		}

		// Update the collection of tag objects
		ConvertTextIntoTags();

		// Recalculate window size based on the
		// current tags collection
		RecalculateWindowSize();

		bool bSourceChanged = m_bSourceChanged;

		m_bSourceChanged = false;
		m_bRecalculating = false;

		Fire_OnRecalculate(FALSE, bSourceChanged);

		*success = TRUE;
	}
	else
		AddException(EX_RECURSIVECALL);

	VXPLIB_END
}

STDMETHODIMP CVXPQuickHTML::Draw(long dc, short x, short y, VARIANT_BOOL * success)
{
	VXPLIB_BEGIN

	*success = FALSE;

	DWORD obj_type = ::GetObjectType((HGDIOBJ)dc);
	if(obj_type == OBJ_DC || obj_type == OBJ_MEMDC)
	{

		Recalculate();

		if(m_bTransparent)
		{
			ResetState();
			DrawTags((HDC)dc, x, y, true);
		}
		else
		{
			if(!m_MemoryBMP) // This is when Transparent changes from True to False
				UpdateBitmap();

			HDC memDC = CreateCompatibleDC((HDC)dc);
			HGDIOBJ old = SelectObject(memDC, m_MemoryBMP);
			BitBlt((HDC)dc, x, y, m_size.cx, m_size.cy, memDC, 0, 0, SRCCOPY);
			SelectObject(memDC, old);
			DeleteDC(memDC);
		}
		*success = TRUE;
	}
	else
		AddException(EX_INVALIDHANDLE);

	VXPLIB_END
}

BOOL CALLBACK CVXPQuickHTML::DrawStateProc(HDC hdc, LPARAM lData, WPARAM wData, int cx, int cy)
{
	CVXPQuickHTML * pObject = (CVXPQuickHTML*)lData;

	try
	{
		if(pObject->m_bTransparent)
		{
			pObject->ResetState();
			pObject->DrawTags(hdc, 0, 0, true);
		}
		else
		{
			if(!pObject->m_MemoryBMP) // This is when Transparent changes from True to False
				pObject->UpdateBitmap();

			HDC memDC = CreateCompatibleDC(hdc);
			HGDIOBJ old = SelectObject(memDC, pObject->m_MemoryBMP);
			BitBlt(hdc, 0, 0, pObject->m_size.cx, pObject->m_size.cy, memDC, 0, 0, SRCCOPY);
			SelectObject(memDC, old);
			DeleteDC(memDC);
		}
		return TRUE;
	}
	catch(...)
	{
	}
	return FALSE;
}

STDMETHODIMP CVXPQuickHTML::DrawState(long dc, short x, short y, long dssFlags, OLE_COLOR Color, VARIANT_BOOL * success)
{
	VXPLIB_BEGIN

	*success = FALSE;

	DWORD obj_type = ::GetObjectType((HGDIOBJ)dc);
	if(obj_type == OBJ_DC || obj_type == OBJ_MEMDC)
	{
		Recalculate();

		HBRUSH hBrush = (dssFlags&dssMono)?::CreateSolidBrush(Color):NULL;
		::DrawState((HDC)dc, hBrush, DrawStateProc, (long)this, 0, x, y, m_size.cx, m_size.cy, dssFlags|DST_COMPLEX);
		if(hBrush)
			::DeleteObject(hBrush);

		*success = TRUE;
	}
	else
		AddException(EX_INVALIDHANDLE);

	VXPLIB_END
}

// ****************************************************
// *** Private methods ********************************
// ****************************************************

void CVXPQuickHTML::ConvertTextIntoTags()
{
	m_pLastTag = NULL;
	DELETE_COLLECTION(m_pTags)
	CString tmp = m_text;

	while(m_pLastTag = GetNextTagObject(tmp))
		ADD_COLLECTION(m_pTags, m_pLastTag)
}

// This function recalculates the tooltip window size
// based on the current collection of tags, and sets the
// value into m_size
void CVXPQuickHTML::RecalculateWindowSize()
{
	if(!m_pTags) // If nothing to calculate at all
	{
		m_size.Zero();	// Set tooltip size to {0,0}
		return;
	}

	HDC dc = CreateCompatibleDC(NULL);
	m_size = m_pTags->CalculateSize(dc);
	DeleteDC(dc);

	if(m_MemoryBMP)
	{
		DeleteObject(m_MemoryBMP);
		m_MemoryBMP = NULL;
	}

	NormalizeSizes();

	if(!m_bTransparent && !m_size.IsEmpty())
		UpdateBitmap();
}

void CVXPQuickHTML::UpdateBitmap()
{
	// Updating the bitmap in memory:
	HDC dc = GetDC(NULL);
	HDC memDC = CreateCompatibleDC(dc);
	m_MemoryBMP = CreateCompatibleBitmap(dc, m_size.cx, m_size.cy);
	HGDIOBJ old = SelectObject(memDC, m_MemoryBMP);
	ResetState();
// *** Filling in the bitmap with default background first ***
	COLORREF bgColor;
	m_pSettings->get_bgColor(&bgColor);
	COLORREF gfColor;
	m_pSettings->get_gfColor(&gfColor);
	VARIANT_BOOL gfEnabled;
	m_pSettings->get_gfEnabled(&gfEnabled);
	VARIANT_BOOL gfHorizontal;
	m_pSettings->get_gfHorizontal(&gfHorizontal);
	RECT r = {0, 0, m_size.cx, m_size.cy};
	if(gfEnabled)
		FillGradient(memDC, bgColor, gfColor, r.left, r.top, r.right, r.bottom, gfHorizontal?true:false);
	else
	{
		HBRUSH brush = CreateSolidBrush(bgColor);
		FillRect(memDC, &r, brush);
		DeleteObject(brush);
	}

// *** Here we must draw the BODY background picture,
// *** if such specified:

	if(m_pBackground)
		m_pBackground->Populate(memDC, r.left, r.top, r.right, r.bottom);

// ***********************************************************
	DrawTags(memDC, 0, 0);
	SelectObject(memDC, old);
	DeleteDC(memDC);
	ReleaseDC(NULL, dc);
}

// This function must extract the next tag from the text
// and return an appropriate object. When no more tags found
// it must return NULL.
// The text parameter is cut by the returned length through
// the invocation of the GetNextPhrase() method.

struct CTagElement
{
	LPCTSTR m_TagName;
	short m_TagID;
};

#define TID_B		1
#define TID_FONT	2
#define TID_TABLE	3
#define TID_S		4
#define TID_U		5
#define TID_TR		6
#define TID_TD		7
#define TID_BR		8
#define TID_DOC		9
#define TID_IMG		10
#define TID_BODY	11
#define TID_HR		12
#define TID_TXT		13
#define TID_I		14
#define TID_BGSOUND	15
#define TID_FONT_END	16
#define TID_TABLE_END	17
#define TID_S_END		18
#define TID_U_END		19
#define TID_TR_END		20
#define TID_TD_END		21
#define TID_I_END		22
#define TID_B_END		23
#define TID_CONTROL		24

CTagElement g_TagElements[] = {{_T("/B"), TID_B_END}, {_T("/FONT"), TID_FONT_END},
								{_T("/I"), TID_I_END}, {_T("/S"), TID_S_END},
								{_T("/STRIKE"), TID_S_END}, {_T("/TABLE"), TID_TABLE_END},
								{_T("/TD"), TID_TD_END}, {_T("/TR"), TID_TR_END},
								{_T("/U"), TID_U_END}, {_T("B"), TID_B},
								{_T("BGSOUND"), TID_BGSOUND}, {_T("BODY"), TID_BODY},
								{_T("BR"), TID_BR}, {_T("CONTROL"), TID_CONTROL},
								{_T("DOC"), TID_DOC}, {_T("FONT"), TID_FONT},
								{_T("HR"), TID_HR}, {_T("I"), TID_I},
								{_T("IMG"), TID_IMG}, {_T("S"), TID_S},
								{_T("STRIKE"), TID_S}, {_T("TABLE"), TID_TABLE},
								{_T("TD"), TID_TD}, {_T("TR"), TID_TR},
								{_T("TXT"), TID_TXT}, {_T("U"), TID_U}};

CTagObject * CVXPQuickHTML::GetNextTagObject(CString & text)
{
	CString phrase;
	while(text.GetLength())
	{
		bool isText;
		if(!GetNextPhrase(text, phrase, isText))
			return NULL;

//****************************************************************
		if(isText)
		{
			if(text.IsEmpty())		// If this is the last object
				phrase.TrimRight();	// Remove all last spaces

			// This variable will be set True, if this text object must have all its
			// preceding spaces removed.
			bool bCutSpaces = false;
			if(m_pLastTag)
			{
				int attr = m_pLastTag->GetTagAttr();
				if(attr != TA_TEXT && attr != TA_COMMENT && attr != TA_SOUND)
					bCutSpaces = true;
				else
				{
					// We can add spaces only after a non-empty text object
					CTagObject * pObject = m_pLastTag;
					while(pObject)
					{
						attr = pObject->GetTagAttr();
						// All these three objects have method GetText():
						if(attr != TA_TEXT && attr != TA_COMMENT && attr != TA_SOUND)
						{
							bCutSpaces = true;
							break;
						}
						if(!pObject->GetText().IsEmpty())
							break;
						pObject = (CTagObject*)pObject->Prev;
					}
					if(!pObject)
						bCutSpaces = true;
				}
			}
			else
				bCutSpaces = true;
			return new CTOText(phrase, this, bCutSpaces);
		}
//****************************************************************
		
		// Otherwise we'll try to identify the tag
		CTag tag(phrase);
		CString name = tag.GetTagName();

		if(tag.GetAttr() != TA_TEXT) // Comment block found
		{
			if(tag.GetAttr() == TA_COMMENT)
				return new CTOComment(name);
			else
				return new CTOSound(name);
		}
		
		// Fastest possible search:
		int index = sizeof(g_TagElements)/sizeof(CTagElement) / 2;
		float step = (float)(sizeof(g_TagElements)/sizeof(CTagElement)) / 4;
		int TagID = -1;
		
		while(step >= 0.25)
		{
			int cmp = name.Compare(g_TagElements[index].m_TagName);

			if(!cmp) // If found the color
			{
				TagID = g_TagElements[index].m_TagID;
				break;
			}
			if(cmp < 0)
				index -= Int(step);
			else
				index += Int(step);

			step /= 2;
		}

		if(TagID == -1)
			continue;

		switch(TagID)
		{
		case TID_TABLE:		return new CTOTable(&tag, this);
		case TID_TABLE_END:	return new CTOTableEnd(&tag);
		case TID_TR:		return new CTOTableTR(&tag, this);
		case TID_TR_END:	return new CTOTableTREnd(&tag);
		case TID_TD:		return new CTOTableTD(&tag, this);
		case TID_TD_END:	return new CTOTableTDEnd(&tag);
		case TID_HR:		return new CTO_HR(&tag, this);
		case TID_BR:		return new CTO_BR(&tag, this);
		case TID_FONT:		return new CTOFont(&tag, this);
		case TID_FONT_END:	return new CTOFontEnd(this);
		case TID_B:			return new CTOBold(&tag, this);
		case TID_B_END:		return new CTOBoldEnd(this);
		case TID_I:			return new CTOItalic(&tag, this);
		case TID_I_END:		return new CTOItalicEnd(this);
		case TID_U:			return new CTOUnderline(&tag, this);
		case TID_U_END:		return new CTOUnderlineEnd(this);
		case TID_S:			return new CTOStrike(&tag, this);
		case TID_S_END:		return new CTOStrikeEnd(this);
		case TID_IMG:		return new CTOImage(&tag, this);
		case TID_DOC:		return new CTODocument(&tag, this);
		case TID_CONTROL:	return new CTOControl(&tag, this);
		case TID_BODY:
			{
				SetupBody(&tag);
				break;;
			}
		case TID_BGSOUND:
			{
				SetupSound(&tag);
				break;
			}
		case TID_TXT:
			{
				CTOTxt TxtObject(&tag, this);
				switch(TxtObject.m_nFormat)
				{
				case TF_HTML:
					{
						if(TxtObject.m_text.IsEmpty())
							continue;
						CString tmp = text;
						text = TxtObject.m_text;
						text += tmp;
						break;
					}
				case TF_NONE:
				case TF_SIMPLE:
					{
						if(TxtObject.m_pList)
							return TxtObject.m_pList;
						break;
					}
				default:
					break;
				}
				break;
			}
		default:
			break;
		}
	}
	return NULL;
}

// It draws the tooltip contents
// The function is to ignore all settings of <BODY>
void CVXPQuickHTML::DrawTags(HDC dc, int x, int y, bool Transparent)
{
	if(m_pTags)
	{

		if(Transparent)
		{
			HDC MemDC = CreateCompatibleDC((HDC)dc);
			int SavedMemDC = SaveDC(MemDC);
			HBITMAP MemBmp = CreateCompatibleBitmap((HDC)dc, m_size.cx, m_size.cy);
			HGDIOBJ OldObject = SelectObject(MemDC, MemBmp);
			BitBlt(MemDC, 0, 0, m_size.cx, m_size.cy, (HDC)dc, x, y, SRCCOPY);

			m_pTags->Draw(MemDC, 0, 0);

			BitBlt((HDC)dc, x, y, m_size.cx, m_size.cy, MemDC, 0, 0, SRCCOPY);
			SelectObject(MemDC, OldObject);
			RestoreDC(MemDC, SavedMemDC);
			DeleteObject(MemBmp);
			DeleteDC(MemDC);
		}
		else

			m_pTags->Draw(dc, x, y);
	}
}

// This function reads text for the next block of information,
// which is whether a tag or a text. When it is a text,
// isText is set to true.
// The 'text' is cut to the next symbol
// If found, the phrase is set to the text found, and true is returned
// when a tag is found
bool CVXPQuickHTML::GetNextPhrase(CString & text, CString & phrase, bool & isText)
{
	if(text.IsEmpty())
		return false;
	int L = text.GetLength();
	int tag_index = text.Find('<');
	if(tag_index == -1)
	{
		if(L)
		{
			phrase = text;
			text.Empty();
			isText = true;
			return true;
		}
		text.Empty();
		return false;
	}
	if(tag_index > 0) // Means we found a simple text
	{
		phrase = text.Left(tag_index);
		text.Delete(0, tag_index);
		isText = true;
		return true;
	}

	// Now we know that tag_index = 0, so we do not need it
	tag_index = text.Find('>', 1);
	if(tag_index == -1) // All the rest is not correct at all
	{
		text.Empty();
		return false;
	}
	phrase = text.Mid(1, tag_index - 1);
	text.Delete(0, tag_index + 1);
	isText = false;
	return true;
}

void CVXPQuickHTML::DrawBuffered(HDC dc, short x, short y, short width, short height)
{
	if(m_MemoryBMP)
	{
		HDC memDC = CreateCompatibleDC(dc);
		HGDIOBJ old = SelectObject(memDC, m_MemoryBMP);
		short w = (width<0)?m_size.cx:width;
		short h = (height<0)?m_size.cy:height;
		if(width >= 0 || height >= 0)
			StretchBlt(dc, x, y, w, h, memDC, 0, 0, m_size.cx, m_size.cy, SRCCOPY);
		else
			BitBlt(dc, x, y, w, h, memDC, 0, 0, SRCCOPY);
		SelectObject(memDC, old);
		DeleteDC(memDC);
	}
}

STDMETHODIMP CVXPQuickHTML::get_Transparent(VARIANT_BOOL *pVal)
{
	VXPLIB_BEGIN

	*pVal = m_bTransparent;

	VXPLIB_END
}

STDMETHODIMP CVXPQuickHTML::put_Transparent(VARIANT_BOOL newVal)
{
	VXPLIB_BEGIN

	m_bTransparent = (newVal != 0);

	VXPLIB_END
}

STDMETHODIMP CVXPQuickHTML::get_Width(short *pVal)
{
	VXPLIB_BEGIN

	Recalculate();

	*pVal = m_size.cx;

	VXPLIB_END
}

STDMETHODIMP CVXPQuickHTML::get_Height(short *pVal)
{
	VXPLIB_BEGIN

	Recalculate();

	*pVal = m_size.cy;

	VXPLIB_END
}

// This function must set the default resourse and return true
// into "succeeded", if LoadLibrary succeeded
// Otherwise, the method must throw a COM exception with the error
// explanation: "Failed to set the default resource on..."
STDMETHODIMP CVXPQuickHTML::SetDefaultRes(BSTR FilePath, VARIANT_BOOL * succeeded)
{
	VXPLIB_BEGIN

	CGlobalObject::SetDefaultRes(FilePath, succeeded);

	VXPLIB_END
}

STDMETHODIMP CVXPQuickHTML::AddFilePath(BSTR FilePath, short *counter)
{
	VXPLIB_BEGIN

	*counter = CGlobalObject::AddFilePath(FilePath);

	VXPLIB_END
}

CSize CVXPQuickHTML::GetControlSize(long ID, CString & Name, short Width, short Height, CString & Param)
{
	BSTR name = Name.SysAllocString();
	BSTR par = Param.SysAllocString();
	Fire_OnControlGetSize(ID, name, &Width, &Height, par);
	SysFreeString(name);
	SysFreeString(par);
	return CSize(Width, Height);
}

CSize CVXPQuickHTML::GetDocumentSize(CString & FilePath, CString & Param, short Width, short Height)
{
	BSTR fpath = FilePath.SysAllocString();
	BSTR par = Param.SysAllocString();
	Fire_OnDocumentGetSize(fpath, par, &Width, &Height);
	SysFreeString(fpath);
	SysFreeString(par);
	return CSize(Width, Height);
}

void CVXPQuickHTML::DrawDocument(CString & FilePath, CString & Param, HDC dc, short x, short y, short width, short height)
{
	BSTR fpath = FilePath.SysAllocString();
	BSTR par = Param.SysAllocString();
	Fire_OnDocumentDraw(fpath, par, (long)dc, x, y, width, height);
	SysFreeString(fpath);
	SysFreeString(par);
}

CSize CVXPQuickHTML::GetDefIconSize()
{
	short Width, Height;
	m_pSettings->get_DefIconWidth(&Width);
	m_pSettings->get_DefIconHeight(&Height);
	return CSize(Width, Height);
};

CString CVXPQuickHTML::OnGetText(CString & FilePath, CString & Res, long Size, long ID)
{
	BSTR fpath = FilePath.SysAllocString();
	BSTR res = Res.SysAllocString();
	BSTR Result = NULL;
	CString final;
	Fire_OnGetText(fpath, res, Size, ID, &Result);
	if(Result)
	{
		final = Result;
		SysFreeString(Result);
	}
	SysFreeString(fpath);
	SysFreeString(res);
	return final;
}

STDMETHODIMP CVXPQuickHTML::get_settings(IVXPSettings **pVal)
{
	VXPLIB_BEGIN

	*pVal = m_pSettings;

	// We must increment reference counter to the object
	// passed. The onus is on the client to call Release()
	// when the object is no longer used.
	// (It happens automatically in VB, but must be done
	// manually in C++)
	m_pSettings->AddRef();

	VXPLIB_END
}

VXPSysFont CVXPQuickHTML::ReadSysFont(CString & name)
{
	name.Remove(' ');
	if(name.IsEmpty())
		return (VXPSysFont)0;
	name.MakeUpper();
	if(name == _T("CAPT"))
		return fontCaption;
	if(name == _T("SCAPT"))
		return fontSmallCaption;
	if(name == _T("MENU"))
		return fontMenu;
	if(name == _T("MSG"))
		return fontMessage;
	if(name == _T("TOOL") || name == _T("SBAR"))
		return fontToolTip;
	return (VXPSysFont)0;
}

void CVXPQuickHTML::SetupBody(CTag * tag)
{
	if(m_bSetupBody) // If the <BODY> has already been found
		return;

	m_BodyTag = *tag;

	m_bSetupBody = true;
	bool found;

// Gradient support: *********************************
	m_grad.Init(tag);
	m_pSettings->put_gfHorizontal(m_grad.m_bGFCHorizontal);
	if(m_grad.m_bGFCValid)
		m_pSettings->put_gfColor(m_grad.m_GFC);	
// ***************************************************

	CString SysFont = tag->GetAttrTextValue(_T("FONT"), &found);
	if(found)
	{
		VXPSysFont font = ReadSysFont(SysFont);
		if(font)
		{
			VARIANT_BOOL success;
			m_pFont->SetFromSystemFont(font, &success);
			ResetState();
		}
	}
// *****************************************************************
	int margin = tag->GetAttrIntValue(_T("LM"), &found);
	if(!found)
		margin = tag->GetAttrIntValue(_T("LEFTMARGIN"), &found);
	if(found)
		m_pMargins->put_Left(margin);
// *****************************************************************
	margin = tag->GetAttrIntValue(_T("RM"), &found);
	if(!found)
		margin = tag->GetAttrIntValue(_T("RIGHTMARGIN"), &found);
	if(found)
		m_pMargins->put_Right(margin);
// *****************************************************************
	margin = tag->GetAttrIntValue(_T("TM"), &found);
	if(!found)
		margin = tag->GetAttrIntValue(_T("TOPMARGIN"), &found);
	if(found)
		m_pMargins->put_Top(margin);
// *****************************************************************
	margin = tag->GetAttrIntValue(_T("BM"), &found);
	if(!found)
		margin = tag->GetAttrIntValue(_T("BOTTOMMARGIN"), &found);
	if(found)
		m_pMargins->put_Bottom(margin);
// *****************************************************************
	COLORREF bgColor = tag->GetAttrColorValue(attrBGCOLOR, found);
	if(found)
		m_pSettings->put_bgColor(bgColor);
// *****************************************************************
	COLORREF txtColor = tag->GetAttrColorValue(_T("TEXT"), found);
	if(found)
		m_pFont->put_Color(txtColor);
// ******************************************************************
	
	m_pBackground = CPicture::GetObjectFromTag(tag, NULL, this);
}

void CVXPQuickHTML::SetupSound(CTag * tag)
{
	if(m_bSetupSound) // If the <BGSOUND> has already been found
		return;
	m_bSetupSound = true;

	m_pSoundClass->SetupSound(tag, this);
}

STDMETHODIMP CVXPQuickHTML::DrawStretched(long dc, short x, short y, short width, short height, VARIANT_BOOL * success)
{
	VXPLIB_BEGIN

	*success = FALSE;

	DWORD obj_type = ::GetObjectType((HGDIOBJ)dc);
	if(obj_type == OBJ_DC || obj_type == OBJ_MEMDC)
	{

		// If SetSourceText was called, but RecalculateLayout
		// wasn't called after this.
		Recalculate();

		if(!m_MemoryBMP) // This is when Transparent changes from True to False
			UpdateBitmap();

		HDC memDC = CreateCompatibleDC((HDC)dc);
		HGDIOBJ old = SelectObject(memDC, m_MemoryBMP);
		StretchBlt((HDC)dc, x, y, (width<0)?m_size.cx:width, (height<0)?m_size.cy:height, memDC, 0, 0, m_size.cx, m_size.cy, SRCCOPY);
		SelectObject(memDC, old);
		DeleteDC(memDC);

		*success = TRUE;
	}
	else
		AddException(EX_INVALIDHANDLE);

	VXPLIB_END
}

STDMETHODIMP CVXPQuickHTML::AboutBox()
{
	VXPLIB_BEGIN

	CAboutBox dlg;
	dlg.DoModal();

	VXPLIB_END
}

// Here we must draw the background color, as defined in the settings,
// and background picture, if was specified.
STDMETHODIMP CVXPQuickHTML::DrawBackground(long dc, short x, short y, short width, short height, VARIANT_BOOL DrawBackColor, VARIANT_BOOL *success)
{
	VXPLIB_BEGIN

	DWORD obj_type = ::GetObjectType((HGDIOBJ)dc);
	if(obj_type == OBJ_DC || obj_type == OBJ_MEMDC)
	{
		Recalculate();

		if(DrawBackColor)
		{
			COLORREF gfColor;
			m_pSettings->get_gfColor(&gfColor);
			VARIANT_BOOL gfEnabled;
			m_pSettings->get_gfEnabled(&gfEnabled);
			VARIANT_BOOL gfHorizontal;
			m_pSettings->get_gfHorizontal(&gfHorizontal);
			COLORREF bgColor;
			m_pSettings->get_bgColor(&bgColor);
			RECT r = {x, y, width, height};
			if(gfEnabled)
				FillGradient((HDC)dc, bgColor, gfColor, r.left, r.top, r.right, r.bottom, gfHorizontal?true:false);
			else
			{
				HBRUSH brush = CreateSolidBrush(bgColor);
				FillRect((HDC)dc, &r, brush);
				DeleteObject(brush);
			}
		}

		if(m_pBackground)
		{
			m_pBackground->Populate((HDC)dc, x, y, width, height);
			*success = TRUE;
		}
		else
			*success = FALSE;
	}
	else
		AddException(EX_INVALIDHANDLE);

	VXPLIB_END
}

STDMETHODIMP CVXPQuickHTML::get_HasBackgroundPicture(VARIANT_BOOL *pVal)
{
	VXPLIB_BEGIN

	Recalculate();

	*pVal = m_pBackground?TRUE:FALSE;

	VXPLIB_END
}

STDMETHODIMP CVXPQuickHTML::GetSpeechText(VXPSpeechFormat format, VARIANT_BOOL bIncludeComments, VARIANT_BOOL bIncludeText, BSTR *Text)
{
	VXPLIB_BEGIN

	if(!m_pTags)
	{
		*Text = NULL;
		return S_OK;
	}

	Recalculate();

	CString result;
	CTagObject * A = m_pTags;
	CString tmp;
	bool bPassedVisible = false; // Passed a visible object after text
	while(A)
	{
		if(!A->IsUnused())
		{
			if(A->IsTextSplitter()) // If any visible object except text
				bPassedVisible = true;
			else
				if((bIncludeText && A->GetTagAttr() == TA_TEXT) || (bIncludeComments && A->GetTagAttr() == TA_COMMENT))
				{
					tmp = A->GetText();
					if(!tmp.IsEmpty())
					{
						if(bPassedVisible)
						{
							if(!result.IsEmpty())
								result += _T(" ");
							bPassedVisible = false;
						}
						result += tmp;
					}
				}
				else
					if(A->GetTagAttr() == TA_SOUND)
						switch(format)
						{
						case sfMicrosoftTTS:
							{
								CString str(_T("\\"));
								str += A->GetText();
								str += _T("\\");
								if(str.GetLength() > 2)
									result += str;
								break;
							}
						case sfXML:
							{
								CString str(_T("<"));
								str += A->GetText();
								str += _T(">");
								if(str.GetLength() > 2)
									result += str;
								break;
							}
						case sfDefault:
							{
								result += A->GetText();
								break;
							}
						default:
							break;
						}
		}
		A = (CTagObject*)A->Next;
	}

	*Text = result.SysAllocString();

	VXPLIB_END
}

STDMETHODIMP CVXPQuickHTML::get_BodyCodePage(long *pVal)
{
	VXPLIB_BEGIN

	*pVal = m_BodyCodePage;

	VXPLIB_END
}

STDMETHODIMP CVXPQuickHTML::get_TxtCodePage(long *pVal)
{
	VXPLIB_BEGIN

	*pVal = m_TxtCodePage;

	VXPLIB_END
}

STDMETHODIMP CVXPQuickHTML::put_BodyCodePage(long newVal)
{
	VXPLIB_BEGIN

	if(!IsValidCodePage(newVal) && newVal)
		AddException(EX_UNSUPPORTEDCODEPAGE);
	else
		m_BodyCodePage = newVal;

	VXPLIB_END
}

STDMETHODIMP CVXPQuickHTML::put_TxtCodePage(long newVal)
{
	VXPLIB_BEGIN

	if(!IsValidCodePage(newVal) && newVal)
		AddException(EX_UNSUPPORTEDCODEPAGE);
	else
		m_TxtCodePage = newVal;

	VXPLIB_END
}

// The output result must never exceed 10.000.000 pixels
void CVXPQuickHTML::NormalizeSizes()
{
	if(((long)m_size.cx * (long)m_size.cy) > 10000000)
	{
		AddException(EX_OUTPUTEXCESS);
		m_size.Zero();
	}
	else
		if(!m_size.cx || !m_size.cy)
			m_size.Zero();
}

STDMETHODIMP CVXPQuickHTML::GradientFillRect(long hDC, OLE_COLOR StartColor, OLE_COLOR EndColor, short x1, short y1, short x2, short y2, VARIANT_BOOL bHorizontal, VARIANT_BOOL *success)
{
	VXPLIB_BEGIN

	*success = FALSE;
	*success = FillGradient((HDC)hDC, StartColor, EndColor, x1, y1, x2, y2, bHorizontal?true:false, false);

	VXPLIB_END
}

STDMETHODIMP CVXPQuickHTML::GetBodyAttr(BSTR attrName, enum VXPTagAttribute attrType, VARIANT *Value)
{
	VXPLIB_BEGIN

	Recalculate();

	if(m_bSetupBody)
		*Value = GetTagAttr(&m_BodyTag, attrName, attrType);
	else
	{
		VARIANT empty;
		VariantClearEx(&empty);
		*Value = empty;
	}

	VXPLIB_END
}

STDMETHODIMP CVXPQuickHTML::GetControlAttr(long ID, BSTR attrName, enum VXPTagAttribute attrType, VARIANT *Value)
{
	VXPLIB_BEGIN

	VARIANT empty;
	VariantClearEx(&empty);
	*Value = empty;

	Recalculate();

	CTOControl * pControl = FindControl(ID);
	if(pControl)
		*Value = GetTagAttr(&pControl->m_Tag, attrName, attrType);

	VXPLIB_END
}

STDMETHODIMP CVXPQuickHTML::GetControlAttrIdx(long Index, BSTR attrName, enum VXPTagAttribute attrType, VARIANT *Value)
{
	VXPLIB_BEGIN

	VARIANT empty;
	VariantClearEx(&empty);
	*Value = empty;

	Recalculate();

	CTOControl * pControl = FindControl(Index, true);
	if(pControl)
		*Value = GetTagAttr(&pControl->m_Tag, attrName, attrType);
	else
		AddException(EX_INDEXOUTOFRANGE);

	VXPLIB_END
}

CTOControl * CVXPQuickHTML::FindControl(long ID, bool bIndex)
{
	if(ID || bIndex) // Control with ID 0 cannot be identified, but with index 0 - possibly
	{
		if(bIndex && ID < 0) // No negative indexes
			return NULL;

		if(!bIndex && m_pLastRequestedControl && m_pLastRequestedControl->m_ID == ID)
			return m_pLastRequestedControl;
		CTOControl * pControl = (CTOControl*)m_pTags;
		int I = 0;
		if(bIndex && m_pLastRequestedControl && m_LastRequestedIndex <= ID)
		{
			pControl = m_pLastRequestedControl;
			I = m_LastRequestedIndex;
		}
		while(pControl)
		{
			if(pControl->GetTagAttr() == TA_CONTROL && !pControl->IsUnused())
			{
				if((bIndex && ID == I) || (!bIndex && pControl->m_ID == ID))
				{
					m_pLastRequestedControl = pControl;
					m_LastRequestedIndex = I;
					return pControl;
				}
				I ++;
			}
			pControl = (CTOControl*)pControl->Next;
		}
	}
	return NULL;
}

VARIANT CVXPQuickHTML::GetTagAttr(CTag * pTag, BSTR attrName, enum VXPTagAttribute attrType)
{
	VARIANT result;
	VariantClearEx(&result);

	bool found;

	CString name(attrName);
	name.Remove(' ');	// Name of the attribute cannot contain spaces
	name.MakeUpper();	// Make to upper register before comparing

	CString tmp;

	switch(attrType)
	{
	case taText:
		{
			tmp = pTag->GetAttrTextValue(name, &found);
			if(found)
			{
				result.vt = VT_BSTR;
				result.bstrVal = tmp.SysAllocString();
			}
			break;
		}
	case taByte:
		{
			tmp = pTag->GetAttrTextValue(name, &found);
			if(found)
			{
				result.vt = VT_UI1;
				result.bVal = (BYTE)_ttoi(tmp);
			}
			break;
		}
	case taShort:
		{
			tmp = pTag->GetAttrTextValue(name, &found);
			if(found)
			{
				result.vt = VT_I2;
				result.iVal = _ttoi(tmp);
			}
			break;
		}
	case taLong:
		{
			tmp = pTag->GetAttrTextValue(name, &found);
			if(found)
			{
				result.vt = VT_I4;
				result.lVal = _ttoi(tmp);
			}
			break;
		}
	case taFloat:
		{
			tmp = pTag->GetAttrTextValue(name, &found);
			if(found)
			{
				result.vt = VT_R4;
				TCHAR * StopScan;
				result.fltVal = (float)_tcstod(tmp, &StopScan);
			}
			break;
		}
	case taDouble:
		{
			tmp = pTag->GetAttrTextValue(name, &found);
			if(found)
			{
				result.vt = VT_R8;
				TCHAR * StopScan;
				result.dblVal = _tcstod(tmp, &StopScan);
			}
			break;
		}
	case taColor:
		{
			COLORREF color = pTag->GetAttrColorValue(name, found);
			if(found)
			{
				result.vt = VT_I4;
				result.lVal = color;
			}
			break;
		}
	case taPercent:
		{
			short Percent = pTag->GetAttrPercentValue(name, found);
			if(found)
			{
				result.vt = VT_UI1;
				result.bVal = (BYTE)Percent;
			}
			break;
		}
	case taDateTime:
		{
			tmp = pTag->GetAttrTextValue(name, &found);
			if(found)
			{
				DATE d;
				if(!FAILED(VarDateFromStr((LPOLESTR)T2COLE((LPCTSTR)tmp), LANG_USER_DEFAULT, 0, &d)))
				{
					result.vt = VT_DATE;
					result.date = d;
				}
			}
			break;
		}
	case taBoolean:
		{
			tmp = pTag->GetAttrTextValue(name, &found);
			if(found)
			{
				VARIANT_BOOL b;
				if(!FAILED(VarBoolFromStr((LPOLESTR)T2COLE((LPCTSTR)tmp), LANG_USER_DEFAULT, 0, &b)))
				{
					result.vt = VT_BOOL;
					result.boolVal = b;
				}
			}
			break;
		}

	default:
		break;
	}

	return result;
}

STDMETHODIMP CVXPQuickHTML::GetControlPos(long ID, short *X, short *Y, short *W, short *H, VARIANT_BOOL *bFound)
{
	VXPLIB_BEGIN

	*bFound = FALSE;

	Recalculate();

	CTOControl * pControl = FindControl(ID);
	if(pControl)
	{
		*X = (short)pControl->m_client.left;
		*Y = (short)pControl->m_client.top;
		*W = (short)pControl->m_client.right;
		*H = (short)pControl->m_client.bottom;
		*bFound = TRUE;
	}

	VXPLIB_END
}

void CVXPQuickHTML::VariantClearEx(VARIANT *pVariant)
{
	memset(pVariant, 0, sizeof(VARIANT));
}

STDMETHODIMP CVXPQuickHTML::TextToHTML(BSTR strSource, long nMaxWidth, BSTR * strDest)
{
	VXPLIB_BEGIN

	CString result = strSource;
	*strDest = result.TextToHTML(nMaxWidth).SysAllocString();

	VXPLIB_END
}

STDMETHODIMP CVXPQuickHTML::TextToHTMLEx(BSTR strSource, long nMaxWidth, long *nRows, VARIANT_BOOL *bTruncated, enum VXPTextToHTML tthOptions, BSTR *strDest)
{
	VXPLIB_BEGIN

	CString result = strSource;
	bool bTrunc = (*bTruncated)?true:false;
	*strDest = result.TextToHTMLEx(nMaxWidth, *nRows, bTrunc, tthOptions).SysAllocString();
	*bTruncated = bTrunc?TRUE:FALSE;

	VXPLIB_END
}

// This method was left here for hystorical reasons:
// The component was initially designed not as open-source,
// and required licensing. 
STDMETHODIMP CVXPQuickHTML::GetLicenses(long *nLicenses)
{
	VXPLIB_BEGIN

	*nLicenses = 1;

	VXPLIB_END
}

STDMETHODIMP CVXPQuickHTML::GetControlsInfo(long * MinID, long * MaxID, long * nControls)
{
	VXPLIB_BEGIN

	*nControls = 0;
	
	if(Recalculate() || !m_nControls)
	{
		// Resetting the variables:
		m_nControls = 0;
		m_MinID = 0;
		m_MaxID = 0;

		CTOControl * pControl = (CTOControl*)m_pTags;
		while(pControl)
		{
			if(pControl->GetTagAttr() == TA_CONTROL && !pControl->IsUnused())
			{
				m_nControls ++;
				if(m_nControls == 1)
				{
					m_MinID = pControl->m_ID;
					m_MaxID = pControl->m_ID;
				}
				else
					if(pControl->m_ID < m_MinID)
						m_MinID = pControl->m_ID;
					else
						if(pControl->m_ID > m_MaxID)
							m_MaxID = pControl->m_ID;
			}
			pControl = (CTOControl*)pControl->Next;
		}
		*nControls = m_nControls;
	}
	else
		*nControls = m_nControls;

	if(m_nControls)
	{
		if(MinID)
			*MinID = m_MinID;
		if(MaxID)
			*MaxID = m_MaxID;
	}

	VXPLIB_END
}

bool CVXPQuickHTML::Recalculate()
{
	if(m_bSourceChanged)
	{
		VARIANT_BOOL success;
		RecalculateLayout(&success);
		return true;
	}
	return false;
}

