/*
 * VXPLib COMPONENT
 *
 * Module: VXPTooltip.cpp - Implementation of class CVXPTooltip
 *
 * Synopsis: Class CVXPTooltip implements methods and properties
 * of interface IVXPTooltip.
 * The actual implementation of the tooltip window is provided
 * by class CToolWnd, while this class merely forwards all its calls
 * to that class.
 *
 * Copyright (c) 2002, Vitaly Tomilov
 * Copyright (c) 2003-2005, Tooltips.Net L.L.C.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * Library support web site - http://www.tooltips.net
 */

#include "stdafx.h"
#include "VXPLib.h"
#include "VXPTooltip.h"

// The following is the list of default margins used for tooltips,
// all set to 2 pixels:
#define DM_REGULARMARGIN	2
#define DM_LEFTMARGIN	DM_REGULARMARGIN
#define DM_TOPMARGIN	DM_REGULARMARGIN
#define DM_RIGHTMARGIN	DM_REGULARMARGIN
#define DM_BOTTOMMARGIN	DM_REGULARMARGIN

/////////////////////////////////////////////////////////////////////////////
// CVXPTooltip

STDMETHODIMP CVXPTooltip::InterfaceSupportsErrorInfo(REFIID riid)
{
	static const IID* arr[] = 
	{
		&IID_IVXPTooltip
	};
	for (int i=0; i < sizeof(arr) / sizeof(arr[0]); i++)
	{
		if (InlineIsEqualGUID(*arr[i],riid))
			return S_OK;
	}
	return S_FALSE;
}

CVXPTooltip::CVXPTooltip()
{
	m_bCalling = false;
}

STDMETHODIMP CVXPTooltip::get_html(IVXPQuickHTML ** pVal)
{
	VXPLIB_BEGIN

	*pVal = m_pHTML;

	// We must increment reference counter to the object
	// passed. The onus is on the client to call Release()
	// when the object is no longer used.
	// (It happens automatically in VB, but must be done
	// manually in C++)
	m_pHTML->AddRef();

	VXPLIB_END
}

HRESULT CVXPTooltip::FinalConstruct()
{
	::CoCreateInstance(CLSID_VXPQuickHTML, NULL, CLSCTX_ALL, IID_IVXPQuickHTML, (void**)&m_pHTML);
	DispEventAdvise(m_pHTML);

	// Calling DispEventAdvise() will increment reference to this object,
	// which we do not need, so we decrement the reference.
	Release();
	
	m_pHTML->get_settings(&m_pSettings);
	m_pSettings->get_margins(&m_pMargins);
	m_pSettings->get_font(&m_pFont);
	m_pSettings->get_bgsound(&m_pSound);

	m_tooltip.Construct(this);

	return S_OK;
}

HRESULT CVXPTooltip::FinalRelease()
{
	m_tooltip.Destruct();

	m_pSound->Release();
	m_pFont->Release();
	m_pMargins->Release();
	m_pSettings->Release();

	// Because DispEventUnadvise() is gonna decrement the reference
	// to this object we must make sure it won't trigger an exception
	// since we decremented the reference in the FinalConstruct(),
	// so we increment the reference before calling DispEventUnadvise()
	AddRef();

	DispEventUnadvise(m_pHTML);

	m_pHTML->Release();

	return S_OK;
}

// Just before this event is fired IVXPQuickHTML calls
// m_pSettings->SetDefaults() for settings to get
// their default values
HRESULT CVXPTooltip::OnSetDefaults()
{
// Changing defaults for the tooltip object:
	VARIANT_BOOL success;
	m_pFont->SetFromSystemFont(fontToolTip, &success);
	m_pFont->put_Color(GetSysColor(COLOR_INFOTEXT));
	m_pSettings->put_bgColor(GetSysColor(COLOR_INFOBK));
	
	m_pMargins->put_Left(DM_LEFTMARGIN);
	m_pMargins->put_Top(DM_TOPMARGIN);
	m_pMargins->put_Right(DM_RIGHTMARGIN);
	m_pMargins->put_Bottom(DM_BOTTOMMARGIN);

	return S_OK;
}

HRESULT CVXPTooltip::OnRecalculate(VARIANT_BOOL Begin, VARIANT_BOOL SourceChanged)
{
	m_tooltip.OnSourceChangedNotify();
	return S_OK;
}

STDMETHODIMP CVXPTooltip::Show(VARIANT_BOOL * success)
{
	VXPLIB_BEGIN

	*success = FALSE;

	if(!m_bCalling)
	{
		m_bCalling = true;
		POINT p;
		GetCursorPos(&p);
		m_tooltip.ShowToolTip((short)p.x, (short)p.y);
		m_bCalling = false;
		*success = TRUE;
	}
	else
		AddException(EX_TOOLRECURSION);
	
	VXPLIB_END
}

STDMETHODIMP CVXPTooltip::ShowAt(short x, short y, VARIANT_BOOL * success)
{
	VXPLIB_BEGIN

	*success = FALSE;

	if(!m_bCalling)
	{
		m_bCalling = true;
		m_tooltip.ShowToolTip(x, y);
		m_bCalling = false;
		*success = TRUE;
	}
	else
		AddException(EX_TOOLRECURSION);
	
	VXPLIB_END
}

STDMETHODIMP CVXPTooltip::Hide(VARIANT_BOOL NoFadingEffect, VARIANT_BOOL * success)
{
	VXPLIB_BEGIN

	*success = FALSE;

	if(!m_bCalling)
	{
		m_bCalling = true;
		m_tooltip.HideToolTip(NoFadingEffect?true:false);
		m_bCalling = false;
		*success = TRUE;
	}
	else
		AddException(EX_TOOLRECURSION);
	
	VXPLIB_END
}

STDMETHODIMP CVXPTooltip::get_Autohide(VARIANT_BOOL *pVal)
{
	VXPLIB_BEGIN

	*pVal = m_tooltip.m_bAutohide;
	
	VXPLIB_END
}

STDMETHODIMP CVXPTooltip::put_Autohide(VARIANT_BOOL newVal)
{
	VXPLIB_BEGIN

	m_tooltip.m_bAutohide = (newVal != 0);
	
	VXPLIB_END
}

STDMETHODIMP CVXPTooltip::get_FadeShow(short *pVal)
{
	VXPLIB_BEGIN

	*pVal = m_tooltip.m_FadeShow;
	
	VXPLIB_END
}

STDMETHODIMP CVXPTooltip::put_FadeShow(short newVal)
{
	VXPLIB_BEGIN

	m_tooltip.m_FadeShow = newVal;
	m_tooltip.SetTheme(ttCustom);
	m_tooltip.UpdateLayering();
	
	VXPLIB_END
}

STDMETHODIMP CVXPTooltip::get_FadeHide(short *pVal)
{
	VXPLIB_BEGIN

	*pVal = m_tooltip.m_FadeHide;
	
	VXPLIB_END
}

STDMETHODIMP CVXPTooltip::put_FadeHide(short newVal)
{
	VXPLIB_BEGIN

	m_tooltip.m_FadeHide = newVal;
	m_tooltip.SetTheme(ttCustom);
	m_tooltip.UpdateLayering();
	
	VXPLIB_END
}

STDMETHODIMP CVXPTooltip::get_Transparency(BYTE *pVal)
{
	VXPLIB_BEGIN

	*pVal = m_tooltip.m_Transparency;
	
	VXPLIB_END
}

STDMETHODIMP CVXPTooltip::put_Transparency(BYTE newVal)
{
	VXPLIB_BEGIN

	m_tooltip.m_Transparency = newVal;
	m_tooltip.SetTheme(ttCustom);
	m_tooltip.UpdateLayering();
	
	VXPLIB_END
}

STDMETHODIMP CVXPTooltip::get_HasTail(VARIANT_BOOL *pVal)
{
	VXPLIB_BEGIN

	*pVal = m_tooltip.m_bHasTail;
	
	VXPLIB_END
}

STDMETHODIMP CVXPTooltip::put_HasTail(VARIANT_BOOL newVal)
{
	VXPLIB_BEGIN

	m_tooltip.m_bHasTail = newVal?true:false;
	m_tooltip.m_TailHeight = newVal?TP_HEIGHT:0;
	m_tooltip.SetTheme(ttCustom);
	
	VXPLIB_END
}

STDMETHODIMP CVXPTooltip::get_SyncWnd(long *pVal)
{
	VXPLIB_BEGIN

	*pVal = (long)m_tooltip.m_hSyncWnd;
	
	VXPLIB_END
}

STDMETHODIMP CVXPTooltip::put_SyncWnd(long newVal)
{
	VXPLIB_BEGIN

	if(!newVal || ::IsWindow((HWND)newVal))
		m_tooltip.m_hSyncWnd = (HWND)newVal;
	else
		AddException(EX_INVALIDHANDLE);

	VXPLIB_END
}

STDMETHODIMP CVXPTooltip::get_Round(BYTE *pVal)
{
	VXPLIB_BEGIN

	*pVal = m_tooltip.m_Round;
	
	VXPLIB_END
}

STDMETHODIMP CVXPTooltip::put_Round(BYTE newVal)
{
	VXPLIB_BEGIN

	m_tooltip.m_Round = newVal;
	m_tooltip.SetTheme(ttCustom);
	
	VXPLIB_END
}

STDMETHODIMP CVXPTooltip::get_State(VXPTooltipState *pVal)
{
	VXPLIB_BEGIN

	*pVal = m_tooltip.m_State;
	
	VXPLIB_END
}

STDMETHODIMP CVXPTooltip::get_HasBorder(VARIANT_BOOL *pVal)
{
	VXPLIB_BEGIN

	*pVal = m_tooltip.m_bHasBorder;
	
	VXPLIB_END
}

STDMETHODIMP CVXPTooltip::put_HasBorder(VARIANT_BOOL newVal)
{
	VXPLIB_BEGIN

	m_tooltip.m_bHasBorder = (newVal != 0);
	m_tooltip.SetTheme(ttCustom);
	
	VXPLIB_END
}

STDMETHODIMP CVXPTooltip::get_BorderColor(OLE_COLOR *pVal)
{
	VXPLIB_BEGIN

	*pVal = m_tooltip.m_BorderColor;
	
	VXPLIB_END
}

STDMETHODIMP CVXPTooltip::put_BorderColor(OLE_COLOR newVal)
{
	VXPLIB_BEGIN

	m_tooltip.m_BorderColor = newVal;
	
	VXPLIB_END
}

STDMETHODIMP CVXPTooltip::get_HasShadow(VARIANT_BOOL *pVal)
{
	VXPLIB_BEGIN

	*pVal = m_tooltip.m_bHasShadow;
	
	VXPLIB_END
}

STDMETHODIMP CVXPTooltip::put_HasShadow(VARIANT_BOOL newVal)
{
	VXPLIB_BEGIN

	m_tooltip.UpdateShadowAttribute(newVal?true:false);
	m_tooltip.SetTheme(ttCustom);
	
	VXPLIB_END
}

STDMETHODIMP CVXPTooltip::get_TraceMode(VARIANT_BOOL *pVal)
{
	VXPLIB_BEGIN

	*pVal = m_tooltip.m_bTraceMode;
	
	VXPLIB_END
}

STDMETHODIMP CVXPTooltip::put_TraceMode(VARIANT_BOOL newVal)
{
	VXPLIB_BEGIN

	m_tooltip.m_bTraceMode = newVal?true:false;
	
	VXPLIB_END
}

STDMETHODIMP CVXPTooltip::get_ShowWhenEmpty(VARIANT_BOOL *pVal)
{
	VXPLIB_BEGIN

	*pVal = m_tooltip.m_bShowWhenEmpty;
	
	VXPLIB_END
}

STDMETHODIMP CVXPTooltip::put_ShowWhenEmpty(VARIANT_BOOL newVal)
{
	VXPLIB_BEGIN

	m_tooltip.m_bShowWhenEmpty = newVal?true:false;
	
	VXPLIB_END
}

STDMETHODIMP CVXPTooltip::get_hWnd(long *pVal)
{
	VXPLIB_BEGIN

	*pVal = (long)m_tooltip.m_hWnd;

	VXPLIB_END
}

STDMETHODIMP CVXPTooltip::get_FadeSound(short *pVal)
{
	VXPLIB_BEGIN

	*pVal = m_tooltip.m_FadeSound;
	
	VXPLIB_END
}

STDMETHODIMP CVXPTooltip::put_FadeSound(short newVal)
{
	VXPLIB_BEGIN

	m_tooltip.m_FadeSound = newVal;
	
	VXPLIB_END
}

STDMETHODIMP CVXPTooltip::get_Theme(enum VXPTooltipTheme *pVal)
{
	VXPLIB_BEGIN

	*pVal = m_tooltip.m_Theme;
	
	VXPLIB_END
}

STDMETHODIMP CVXPTooltip::put_Theme(enum VXPTooltipTheme newVal)
{
	VXPLIB_BEGIN

	m_tooltip.SetTheme(newVal);
	m_tooltip.UpdateLayering();
	
	VXPLIB_END
}

STDMETHODIMP CVXPTooltip::DefWndProc(long uMsg, long wParam, long lParam, long *result)
{
	VXPLIB_BEGIN

	*result = m_tooltip.CallDefWndProc(uMsg, wParam, lParam);
	
	VXPLIB_END
}

STDMETHODIMP CVXPTooltip::Subscribe(long uMsg, VARIANT_BOOL * success)
{
	VXPLIB_BEGIN

	*success = m_tooltip.Subscribe(uMsg);
	
	VXPLIB_END
}

STDMETHODIMP CVXPTooltip::UnSubscribe(long uMsg, VARIANT_BOOL * success)
{
	VXPLIB_BEGIN

	*success = m_tooltip.UnSubscribe(uMsg);
	
	VXPLIB_END
}

STDMETHODIMP CVXPTooltip::Create(VARIANT_BOOL * success)
{
	VXPLIB_BEGIN

	*success = m_tooltip.Create();
	
	VXPLIB_END
}

STDMETHODIMP CVXPTooltip::Destroy(VARIANT_BOOL * success)
{
	VXPLIB_BEGIN

	*success = m_tooltip.Destroy();
	
	VXPLIB_END
}

STDMETHODIMP CVXPTooltip::get_MouseClickResponse(VXPMouseClickResponse * pVal)
{
	VXPLIB_BEGIN

	*pVal = m_tooltip.m_MouseClickResponse;
	
	VXPLIB_END
}

STDMETHODIMP CVXPTooltip::put_MouseClickResponse(VXPMouseClickResponse newVal)
{
	VXPLIB_BEGIN

	m_tooltip.m_MouseClickResponse = newVal;

	VXPLIB_END
}
